import ast
import inspect
import itertools
import logging
from abc import abstractmethod
from typing import Dict, List, Type, Union

import casadi as ca

import pymoca
from pymoca.backends.casadi.model import Model as _Model

from rtctools._internal.alias_tools import AliasDict
from rtctools._internal.caching import cached
from rtctools.optimization.optimization_problem import OptimizationProblem

from . import ConstantInput, ControlInput, Model, SymbolicParameter, Variable

logger = logging.getLogger("mesido")


DYNAMIC_NAME_CACHE = {}


def add_variables_documentation_automatically(class_: Type):
    """
    This function can be added as a decorator to asset classes. It will update the documentation
    of that class with the variables that are created in it and the classes from which it inherits,
    based on the self.add_variable() function. The string
    "{add_variable_names_for_documentation_here}" in the asset documentation is then replaced by
    a list of the variable names.

    Note: This decorator must be added to any class which may be referenced by `class_`.

    Note: This function will not work properly if 2 two port classes have the same name due to
    the use of `DYNAMIC_NAME_CACHE`.

    Args:
        class_: The asset class it is documenting

    Returns: updated documentation of the class.

    """

    def get_names_for_class(current_class_: Type) -> List[str]:
        """
        This function checks for the variables that are added using the add_variable function.
        Using the cache the variable names that are created by inheritance and port names,
        are tracked to recreate the full variable name.

        It will follow any classes up the inheritance tree and it will follow any Port classes
        recursively. The variable name of port classes are added to the variable path while classes
        up the inheritance tree only add their variable paths to the set of this class.

        Args:
            current_class_: The class which is currently checked

        Returns: List of variable names

        """
        dynamic_names = []
        for class__ in inspect.getmro(current_class_):
            if inspect.getmodule(class__).__name__ == "builtins":
                break
            ast_of_init: ast.Module = ast.parse(inspect.getsource(class__))
            for node in ast.walk(ast_of_init):
                # Look for any function calls that fit self.add_variable
                if (
                    isinstance(node, ast.Call)
                    and isinstance(node.func, ast.Attribute)
                    and isinstance(node.func.value, ast.Name)
                    and node.func.value.id == "self"
                    and node.func.attr == "add_variable"
                ):
                    # Now extract the first and second argument
                    # First argument decides if it is a port or variable name
                    # Second argument is the string name
                    if isinstance(node.args[0], ast.Name):
                        port_class_name = node.args[0].id
                        if port_class_name in [
                            "HeatPort",
                            "ElectricityPort",
                            "GasPort",
                            "Primary",
                            "Secondary",
                            "_ATESBaseAsset",
                            "_NonStorageComponent",
                            "_NonStorageComponentSinkType",
                            "_NonStorageComponentSourceType",
                            "_StorageComponent",
                        ]:
                            # Follow the port and retrieve all variable names for that port
                            port_name = node.args[1].value

                            # We leverage Python's import structure here. If there is a reference
                            # to a class in another file, that file will already have imported
                            # the necessary class as well as run this function.
                            dynamic_names_of_port = DYNAMIC_NAME_CACHE[port_class_name]
                            for dynamic_name_of_port in dynamic_names_of_port:
                                dynamic_names.append(f"{port_name}.{dynamic_name_of_port}")
                        elif node.args[0].id == "Variable":
                            # This is a variable for this component, save its name.
                            dynamic_names.append(node.args[1].value)
                        else:
                            raise RuntimeError(f"Unknown case:\n{ast.dump(node)}")
                    else:
                        raise RuntimeError(f"Unknown case:\n{ast.dump(node)}")

        dynamic_names = sorted(dynamic_names)
        if current_class_.__name__ in DYNAMIC_NAME_CACHE:
            raise RuntimeError(
                f"Unsupported situation. Already generated documentation for "
                f"{current_class_.__name__} before. Possibly that 2 classes "
                f"(in different modules) have the same class name? This is not "
                f"supported by this function."
            )
        DYNAMIC_NAME_CACHE[current_class_.__name__] = dynamic_names
        return dynamic_names

    all_dynamic_names = get_names_for_class(class_)

    # Format the dynamic names properly
    formatted_dynamic_names = [
        f"* {{name}}.{dynamic_name}" for dynamic_name in sorted(all_dynamic_names)
    ]

    # Find the indent that should be used
    line_with_hook = next(
        (
            line
            for line in class_.__doc__.splitlines()
            if "{" "add_variable_names_for_documentation_here}" in line
        ),
        None,
    )
    if line_with_hook is None:
        indent = ""
    else:
        indent, _ = line_with_hook.split("{add_variable_names_for_documentation_here}")

        # Insert the dynamic names into the documentation
        class_.__doc__ = class_.__doc__.replace(
            "{add_variable_names_for_documentation_here}",
            f"\n{indent}".join(formatted_dynamic_names),
        )
    return class_


class PyCMLMixin(OptimizationProblem):
    def __init__(self, *args, **kwargs):
        logger.debug("Using pymoca {}.".format(pymoca.__version__))

        pycml_model = self.pycml_model()

        self.__flattened_model = pycml_model.flatten()

        self.__pymoca_model = _Model()
        for v in self.__flattened_model.variables.values():
            if isinstance(v, SymbolicParameter):
                self.__pymoca_model.parameters.append(v)
            elif isinstance(v, (ControlInput, ConstantInput)):
                self.__pymoca_model.inputs.append(v)
            elif isinstance(v, Variable) and v.has_derivative:
                self.__pymoca_model.states.append(v)
                self.__pymoca_model.der_states.append(v.der())
            else:
                self.__pymoca_model.alg_states.append(v)

        self.__pymoca_model.equations = self.__flattened_model.equations
        self.__pymoca_model.initial_equations = self.__flattened_model.initial_equations
        self.__pymoca_model.simplify(self.compiler_options())

        if (
            len(self.__flattened_model.inequalities) > 0
            or len(self.__flattened_model.initial_inequalities) > 0
        ):
            raise NotImplementedError("Inequalities are not supported yet")

        # Note that we do not pass the numeric parameters to the Pymoca model
        # in their entirety. That way we can avoid making useless Variable
        # instances, as the parameters do not appear in any equations anyway.
        self.__parameters = {
            k: v
            for k, v in self.__flattened_model.numeric_parameters.items()
            if not isinstance(v, str)
        }
        self.__string_parameters = {
            k: v for k, v in self.__flattened_model.numeric_parameters.items() if isinstance(v, str)
        }

        # Extract the CasADi MX variables used in the model
        self.__mx = {}
        self.__mx["time"] = [self.__pymoca_model.time]
        self.__mx["states"] = [v.symbol for v in self.__pymoca_model.states]
        self.__mx["derivatives"] = [v.symbol for v in self.__pymoca_model.der_states]
        self.__mx["algebraics"] = [v.symbol for v in self.__pymoca_model.alg_states]
        self.__mx["parameters"] = [v.symbol for v in self.__pymoca_model.parameters]
        self.__mx["string_parameters"] = [
            v.name
            for v in (*self.__pymoca_model.string_parameters, *self.__pymoca_model.string_constants)
        ]
        self.__mx["control_inputs"] = []
        self.__mx["constant_inputs"] = []
        self.__mx["lookup_tables"] = []

        for v in self.__pymoca_model.inputs:
            if v.symbol.name() in self.__pymoca_model.delay_states:
                raise NotImplementedError("Delays are not supported yet")
            else:
                if v.symbol.name() in kwargs.get("lookup_tables", []):
                    raise NotImplementedError()
                elif v.fixed:
                    self.__mx["constant_inputs"].append(v.symbol)
                else:
                    self.__mx["control_inputs"].append(v.symbol)

        # Initialize nominals and types
        # These are not in @cached dictionary properties for backwards compatibility.
        self.__python_types = AliasDict(self.alias_relation)
        for v in itertools.chain(
            self.__pymoca_model.states, self.__pymoca_model.alg_states, self.__pymoca_model.inputs
        ):
            self.__python_types[v.symbol.name()] = v.python_type

        # Initialize dae, initial residuals, as well as delay arguments
        # These are not in @cached dictionary properties so that we need to create the list
        # of function arguments only once.
        variable_lists = ["states", "der_states", "alg_states", "inputs", "constants", "parameters"]
        function_arguments = [self.__pymoca_model.time] + [
            ca.veccat(*[v.symbol for v in getattr(self.__pymoca_model, variable_list)])
            for variable_list in variable_lists
        ]

        self.__dae_residual = self.__pymoca_model.dae_residual_function(*function_arguments)
        if self.__dae_residual is None:
            self.__dae_residual = ca.MX()

        self.__initial_residual = self.__pymoca_model.initial_residual_function(*function_arguments)
        if self.__initial_residual is None:
            self.__initial_residual = ca.MX()

        super().__init__(*args, **kwargs)

    @cached
    def compiler_options(self) -> Dict[str, Union[str, bool]]:
        """
        Subclasses can configure the `pymoca <http://github.com/pymoca/pymoca>`_
        compiler options here.

        :returns: A dictionary of pymoca compiler options. See the pymoca documentation
                  for details.
        """
        compiler_options = {}
        compiler_options["detect_aliases"] = True
        compiler_options["replace_parameter_expressions"] = True
        compiler_options["allow_derivative_aliases"] = False
        return compiler_options

    @property
    def dae_residual(self):
        return self.__dae_residual

    @property
    def dae_variables(self):
        return self.__mx

    @property
    def output_variables(self):
        output_variables = super().output_variables.copy()
        output_variables.extend([ca.MX.sym(variable) for variable in self.__pymoca_model.outputs])
        output_variables.extend(self.__mx["control_inputs"])
        return output_variables

    def parameters(self, ensemble_member):
        parameters = super().parameters(ensemble_member)
        parameters.update(self.__parameters)
        parameters.update({v.name: v.value for v in self.__pymoca_model.parameters})
        return parameters

    def string_parameters(self, ensemble_member):
        parameters = super().string_parameters(ensemble_member)
        parameters.update(self.__string_parameters)
        parameters.update({v.name: v.value for v in self.__pymoca_model.string_parameters})
        parameters.update({v.name: v.value for v in self.__pymoca_model.string_constants})
        return parameters

    @property
    def initial_residual(self):
        return self.__initial_residual

    def bounds(self):
        bounds = super().bounds()

        for v in itertools.chain(
            self.__pymoca_model.states, self.__pymoca_model.alg_states, self.__pymoca_model.inputs
        ):
            sym_name = v.symbol.name()

            try:
                bounds[sym_name] = self.merge_bounds(bounds[sym_name], (v.min, v.max))
            except KeyError:
                if self.__python_types.get(sym_name, float) == bool:
                    bounds[sym_name] = (max(0, v.min), min(1, v.max))
                else:
                    bounds[sym_name] = (v.min, v.max)

        return bounds

    def variable_is_discrete(self, variable):
        return self.__python_types.get(variable, float) != float

    @property
    def alias_relation(self):
        return self.__pymoca_model.alias_relation

    def variable_nominal(self, variable):
        try:
            return self.__nominals[variable]
        except AttributeError:
            self.__nominals = AliasDict(self.alias_relation)

            # Iterate over nominalizable states
            for v in itertools.chain(
                self.__pymoca_model.states,
                self.__pymoca_model.alg_states,
                self.__pymoca_model.inputs,
            ):
                sym_name = v.symbol.name()
                nominal = v.nominal
                if nominal == 0.0:
                    nominal = 1.0
                self.__nominals[sym_name] = nominal

            return self.variable_nominal(variable)
        except KeyError:
            return super().variable_nominal(variable)

    @abstractmethod
    def pycml_model(self) -> Model:
        raise NotImplementedError



from pathlib import Path
from unittest import TestCase

from mesido.esdl.esdl_parser import ESDLFileParser
from mesido.esdl.profile_parser import ProfileReaderFromFile
from mesido.util import run_esdl_mesido_optimization

import numpy as np

from utils_tests import demand_matching_test, energy_conservation_test, heat_to_discharge_test


class TestWarmingUpUnitCases(TestCase):
    def test_1a(self):
        """
        This is the most basic check where we have a simple network and check for the basic physics.
        This simple network includes one source, pipes, a node, and 3 demands.

        Checks;
        - Demand matching
        - Energy conservation
        - Heat to discharge
        - Checks for conservation of flow and heat at the node
        - Check for equal head at all node connections
        - Checks that the minimum pressure-drop constraints at the demand are satisfied
        - Check that Heat_demand & Heat_source are set correctly and are linked to the Heat_flow
        variable

        """
        import models.unit_cases.case_1a.src.run_1a as run_1a
        from models.unit_cases.case_1a.src.run_1a import HeatProblem

        base_folder = Path(run_1a.__file__).resolve().parent.parent

        # Just a "problem is not infeasible"
        heat_problem = run_esdl_mesido_optimization(
            HeatProblem,
            base_folder=base_folder,
            esdl_file_name="1a.esdl",
            esdl_parser=ESDLFileParser,
            profile_reader=ProfileReaderFromFile,
            input_timeseries_file="timeseries_import.xml",
        )

        results = heat_problem.extract_results()

        demand_matching_test(heat_problem, results)
        energy_conservation_test(heat_problem, results)
        heat_to_discharge_test(heat_problem, results)

        for node, connected_pipes in heat_problem.energy_system_topology.nodes.items():
            discharge_sum = 0.0
            heat_sum = 0.0

            for i_conn, (_pipe, orientation) in connected_pipes.items():
                discharge_sum += results[f"{node}.HeatConn[{i_conn+1}].Q"] * orientation
                heat_sum += results[f"{node}.HeatConn[{i_conn+1}].Heat"] * orientation
                np.testing.assert_allclose(
                    results[f"{node}.HeatConn[{i_conn+1}].H"], results[f"{node}.H"], atol=1.0e-6
                )

            np.testing.assert_allclose(discharge_sum, 0.0, atol=1.0e-12)
            np.testing.assert_allclose(0.0, heat_sum, atol=1.0e-6)

        for demand in heat_problem.energy_system_components.get("heat_demand", []):
            np.testing.assert_array_less(
                10.2 - 1.0e-6, results[f"{demand}.HeatIn.H"] - results[f"{demand}.HeatOut.H"]
            )
            np.testing.assert_allclose(
                results[f"{demand}.HeatIn.Heat"] - results[f"{demand}.HeatOut.Heat"],
                results[f"{demand}.Heat_demand"],
                atol=1.0e-6,
            )
            np.testing.assert_allclose(
                results[f"{demand}.Heat_demand"], results[f"{demand}.Heat_flow"], atol=1.0e-6
            )

        for source in heat_problem.energy_system_components.get("heat_source", []):
            np.testing.assert_allclose(
                results[f"{source}.HeatOut.Heat"] - results[f"{source}.HeatIn.Heat"],
                results[f"{source}.Heat_source"],
                atol=1.0e-6,
            )
            np.testing.assert_allclose(
                results[f"{source}.Heat_source"], results[f"{source}.Heat_flow"], atol=1.0e-6
            )

    def test_2a(self):
        """
        This is the most basic check where we have a simple network and check for the basic physics.
        This simple network includes two source, pipes, nodes, and 3 demands.

        Checks;
        - Demand matching
        - Energy conservation
        - Heat to discharge

        """
        import models.unit_cases.case_2a.src.run_2a as run_2a
        from models.unit_cases.case_2a.src.run_2a import HeatProblem

        base_folder = Path(run_2a.__file__).resolve().parent.parent

        # Just a "problem is not infeasible"
        heat_problem = run_esdl_mesido_optimization(
            HeatProblem,
            base_folder=base_folder,
            esdl_file_name="2a.esdl",
            esdl_parser=ESDLFileParser,
            profile_reader=ProfileReaderFromFile,
            input_timeseries_file="timeseries_import.xml",
        )

        demand_matching_test(heat_problem, heat_problem.extract_results())
        energy_conservation_test(heat_problem, heat_problem.extract_results())
        heat_to_discharge_test(heat_problem, heat_problem.extract_results())

    def test_3a(self):
        """
        This is the most basic check where we have a simple network and check for the basic physics.
        This simple network includes one source, pipes, node, a tank storage and 3 demands.

        Checks;
        - Demand matching
        - Energy conservation
        - Heat to discharge
        - Check the flow direction variable (this problem should always have a switching flow
        direction for the pipe connected to the buffer tank)
        - Check that the Heat_buffer & Heat_flow variable are set correctly
        - Check that the history for the buffer is set correctly at t=0
        - Check that the heat loss is positive and as expected
        - Check that the Stored heat is the sum of (dis)charge and losses

        """
        import models.unit_cases.case_3a.src.run_3a as run_3a
        from models.unit_cases.case_3a.src.run_3a import HeatProblem

        base_folder = Path(run_3a.__file__).resolve().parent.parent

        # Just a "problem is not infeasible"
        heat_problem = run_esdl_mesido_optimization(
            HeatProblem,
            base_folder=base_folder,
            esdl_file_name="3a.esdl",
            esdl_parser=ESDLFileParser,
            profile_reader=ProfileReaderFromFile,
            input_timeseries_file="timeseries_import.xml",
        )

        results = heat_problem.extract_results()
        parameters = heat_problem.parameters(0)
        bounds = heat_problem.bounds()

        demand_matching_test(heat_problem, results)
        energy_conservation_test(heat_problem, results)
        heat_to_discharge_test(heat_problem, results)

        # We only check the flow directions for the time-steps that there is flow in the pipe.
        inds = np.round(1 - results["Pipe_e53a__is_disconnected"]).astype(bool)
        np.testing.assert_allclose(
            np.round(results["Pipe_e53a__flow_direct_var"][inds]) * 2.0 - 1.0,
            np.sign(results["Pipe_e53a.Q"][inds]),
        )

        for buffer in heat_problem.energy_system_components.get("heat_buffer", []):
            np.testing.assert_allclose(
                results[f"{buffer}.Heat_buffer"], results[f"{buffer}.Heat_flow"]
            )
            np.testing.assert_allclose(
                results[f"{buffer}.HeatIn.Heat"] - results[f"{buffer}.HeatOut.Heat"],
                results[f"{buffer}.Heat_buffer"],
            )
            # buffer should have positive heat loss
            assert parameters[f"{buffer}.heat_loss_coeff"] > 0.0
            np.testing.assert_allclose(
                results[f"{buffer}.Stored_heat"] * parameters[f"{buffer}.heat_loss_coeff"],
                results[f"{buffer}.Heat_loss"],
            )
            np.testing.assert_allclose(
                results[f"{buffer}.Stored_heat"][0], bounds[f"{buffer}.Stored_heat"][0].values[0]
            )
            np.testing.assert_allclose(
                results[f"{buffer}.Stored_heat"][-1] - results[f"{buffer}.Stored_heat"][0],
                np.sum(results[f"{buffer}.Heat_buffer"][1:] * 3600.0)
                - np.sum(results[f"{buffer}.Heat_loss"][1:] * 3600.0),
                atol=1.0,
            )
            np.testing.assert_allclose(results[f"{buffer}.Heat_buffer"][0], 0.0, atol=1.0e-6)

            np.testing.assert_allclose(
                results[f"{buffer}.dH"][inds],
                results[f"{buffer}.HeatOut.H"][inds] - results[f"{buffer}.HeatIn.H"][inds],
                atol=1.0e-6,
            )


if __name__ == "__main__":

    a = TestWarmingUpUnitCases()
    a.test_1a()
    a.test_2a()
    a.test_3a()



import locale
import logging
import time
from pathlib import Path

import esdl

from mesido.esdl.esdl_mixin import DBAccesType
from mesido.esdl.esdl_mixin import ESDLMixin
from mesido.esdl.esdl_parser import ESDLFileParser
from mesido.esdl.profile_parser import ProfileReaderFromFile
from mesido.head_loss_class import HeadLossOption
from mesido.network_common import NetworkSettings
from mesido.physics_mixin import PhysicsMixin
from mesido.workflows.io.write_output import ScenarioOutput
from mesido.workflows.utils.helpers import main_decorator

import numpy as np

from rtctools.optimization.collocated_integrated_optimization_problem import (
    CollocatedIntegratedOptimizationProblem,
)
from rtctools.optimization.goal_programming_mixin import Goal
from rtctools.optimization.linearized_order_goal_programming_mixin import (
    LinearizedOrderGoalProgrammingMixin,
)
from rtctools.optimization.single_pass_goal_programming_mixin import (
    CachingQPSol,
    SinglePassGoalProgrammingMixin,
)
from rtctools.optimization.timeseries import Timeseries
from rtctools.util import run_optimization_problem

DB_HOST = "172.17.0.2"
DB_PORT = 8086
DB_NAME = "Warmtenetten"
DB_USER = "admin"
DB_PASSWORD = "admin"

logger = logging.getLogger("mesido")
logger.setLevel(logging.INFO)

locale.setlocale(locale.LC_ALL, "")

WATT_TO_MEGA_WATT = 1.0e6
WATT_TO_KILO_WATT = 1.0e3


def _extract_values_timeseries(v, type_r=None):
    if isinstance(v, Timeseries):
        v = v.values
        if type_r == "min":
            v = min(v)
        elif type_r == "max":
            v = max(v)
    return v


class OptimisationOverview:
    """
    This class is used to combine the optimisation results, bounds, parameters and aliases
    needed for post-processing and visualisation.
    """

    def __init__(self, total_results, bounds, parameters, aliases):
        self.results = total_results
        self.bounds = bounds
        self.parameters = parameters
        self.aliases = aliases


# -------------------------------------------------------------------------------------------------
# Step 1:
# Match the target demand specified
class TargetDemandGoal(Goal):
    def __init__(self, state, target, priority=1, order=2):
        self.state = state

        self.target_min = target
        self.target_max = target
        self.function_range = (-2.0 * max(target.values), 2.0 * max(target.values))
        self.function_nominal = np.median(target.values)
        self.priority = priority
        self.order = order

    def function(self, optimization_problem, ensemble_member):
        return optimization_problem.state(self.state)


# -------------------------------------------------------------------------------------------------
# Step 2:
# Match the maximum producer profiles
class TargetProducerGoal(Goal):
    def __init__(self, state, target, priority=20, order=2):
        self.state = state

        self.target_min = target
        self.target_max = target
        self.function_range = (-2.0 * max(target.values), 2.0 * max(target.values))
        self.function_nominal = np.median(target.values)
        self.priority = priority
        self.order = order

    def function(self, optimization_problem, ensemble_member):
        return optimization_problem.state(self.state)


# -------------------------------------------------------------------------------------------------
# Step 3:
# The merit order of the sources and consumers (something like [3, 1, 2]), determine the priority
# in which the source is available for use and the consumer to consume.
# The priorities are determined based on the marginal costs provided in ESDL. For producer, the
# lower the marginal costs, the more the producer should be used (e.g. cheap to use). For consumers
# the opposite is true, the higher the marginal costs (e.g. more revenue), the more energy should
# be delivered to that consumer.
# Minimize the source use with lowest priority 3 (highest marginal costs), then the source with
# priority 2 etc, while the consumers are maximised starting with the highest marginal costs.
class MinimizeSourcesGoalMerit(Goal):
    """
    Apply constraints to enforce esdl specified milp producer merit order usage
    """

    def __init__(self, source_variable, prod_priority, func_range_bound, nominal, order=2):
        self.target_max = 0.0  # func_range_bound[0]
        self.function_range = func_range_bound
        self.source_variable = source_variable
        self.function_nominal = nominal
        self.priority = prod_priority
        self.order = order

    def function(self, optimization_problem, ensemble_member):
        return optimization_problem.state(f"{self.source_variable}")


class MaximizeDemandGoalMerit(Goal):
    """
    Apply constraints to enforce maximisation of consumption, priority is based on the marginal
    costs
    """

    def __init__(self, demand_variable, prod_priority, func_range_bound, nominal, order=2):

        self.demand_variable = demand_variable
        self.function_nominal = nominal
        self.priority = prod_priority
        self.order = 1

    def function(self, optimization_problem, ensemble_member):
        return -optimization_problem.state(f"{self.demand_variable}")


class MinimizeStorageGoalMerit(Goal):
    """
    Apply constraints to enforce minimisation of charging of storage, priority is based on the
    marginal costs
    """

    def __init__(self, source_variable, prod_priority, func_range_bound, nominal, order=2):
        self.target_max = 0.0
        self.function_range = func_range_bound
        self.source_variable = source_variable
        self.function_nominal = nominal
        self.priority = prod_priority
        self.order = order

    def function(self, optimization_problem, ensemble_member):
        return optimization_problem.state(f"{self.source_variable}")


class MaximizeStorageGoalMerit(Goal):
    """
    Apply constraints to enforce maximisation of discharging of storage, priority is based on the
    marginal costs
    """

    def __init__(self, demand_variable, prod_priority, func_range_bound, nominal, order=2):
        self.target_min = func_range_bound[1]
        self.function_range = func_range_bound
        self.demand_variable = demand_variable
        self.function_nominal = nominal
        self.priority = prod_priority
        self.order = order

    def function(self, optimization_problem, ensemble_member):
        return optimization_problem.state(f"{self.demand_variable}")


# -------------------------------------------------------------------------------------------------
class _GoalsAndOptions:
    def path_goals(self):
        goals = super().path_goals().copy()

        # Demand and producer matching
        map_demand = {
            "electricity_demand": {
                "target": "target_electricity_demand",
                "state": "Electricity_demand",
            },
            "gas_demand": {"target": "target_gas_demand", "state": "Gas_demand_mass_flow"},
        }
        map_prod = {
            "electricity_source": {
                "target": "maximum_electricity_source",
                "state": "Electricity_source",
            },
            "gas_source": {"target": "maximum_gas_source", "state": "Gas_source_mass_flow"},
        }
        for type, type_values in [*map_demand.items(), *map_prod.items()]:
            for asset in self.energy_system_components.get(type, []):
                timeseries_name = f"{asset}.{type_values['target']}"
                if timeseries_name in self.io.get_timeseries_names():
                    target = self.get_timeseries(timeseries_name)
                    state = f"{asset}.{type_values['state']}"
                    if type in map_demand.keys():
                        goals.append(TargetDemandGoal(state, target))
                    else:
                        priority = 2 * len(self._esdl_assets)
                        goals.append(TargetProducerGoal(state, target, priority))

        return goals


# -------------------------------------------------------------------------------------------------
class MultiCommoditySimulator(
    ScenarioOutput,
    _GoalsAndOptions,
    PhysicsMixin,
    LinearizedOrderGoalProgrammingMixin,
    SinglePassGoalProgrammingMixin,
    ESDLMixin,
    CollocatedIntegratedOptimizationProblem,
):
    """
    This workflow allows for the simulation of (combined) hydrogen and electricity networks,
    containing consumers, producers and conversion assets.
    The priority of the consumers, producers and conversion assets is set using the marginal costs
    in the ESDL file, allowing for flexible customised operation. Producers with the lowest marginal
    costs are maximised in operation before other consumers are used, while consumers with the
    highest marginal costs are maximised before other consumers are satisfied. In case both profiles
    and marginal costs are provided for producers or consumers, then the marginal costs are ignored
    and the profiles will be matched. always are preferred over the
    To obtain this workflow the objective functions are setup according to the scheme described
    below.

    Goal priorities are:
    1. Match target demand specified
    2. Producers with highest marginal costs are minimised and Consumers with highest marginal
    costs are maximised first, working back towards the lower marginal costs.
    3. Producers with a production profile will be matched at end. This goal should be obsolete, as
     the bound is set towards the production profile and other producers are first minimised.

    Notes:
    - Currently all demand profiles can be used, however the length of the simulation is based on
    the length and timestep of these profiles. Too long timehorizons might results in too big
    problems for the solver.
    - No cyclic constraints are yet applied to storages as this workflow solely functions as a
    simulator.
    - TODO: When the number of assets become larger, the simulator might be applied in stages with
    consecutive parts of the time horizon.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._qpsol = None
        self._priorities_output = []

    def pre(self):
        self._qpsol = CachingQPSol()

        super().pre()

    @property
    def esdl_assets(self):
        assets = super().esdl_assets

        for asset in assets.values():
            # Overwrite all assets marked as optional to be used
            if asset.attributes["state"].name in ["OPTIONAL"]:
                asset.attributes["state"] = esdl.AssetStateEnum.ENABLED
                logger.warning(
                    "The following asset has been specified as OPTIONAL but it has been changed "
                    f"to be included in the simulation, asset type: {asset.asset_type }, asset "
                    f"name: {asset.name}"
                )

        return assets

    def __create_asset_list_controls(self, asset_types_to_include, assets_without_control):
        """
        This function creates the lists and dictionaries of assets to include in the optimization
        based on the marginal costs, e.g. priorities. It excludes the assets who already have
        assigned timeseries for demand or production.
        Furthermore, it creates a map of the variables that are required for every asset.
        :param asset_types_to_include:
        :param assets_without_control:
        :return:
        """
        type_variable_map = {
            "electricity_demand": "Electricity_demand",
            "electricity_source": "Electricity_source",
            "gas_demand": "Gas_demand_mass_flow",
            "gas_source": "Gas_source_mass_flow",
            "gas_tank_storage": {"charge": "Gas_tank_flow", "discharge": "__Q_discharge"},
            "electricity_storage": {
                "charge": "Power_charging",
                "discharge": ".Power_discharging",
            },
            "electrolyzer": "Power_consumed",
        }

        assets_to_include = {}
        asset_variable_map = {}
        available_timeseries = [t.split(".")[0] for t in self.io.get_timeseries_names()]
        for group, asset_types in asset_types_to_include.items():
            for asset_type in asset_types:
                for asset in self.energy_system_components.get(asset_type, []):
                    if asset not in available_timeseries:
                        if group in assets_to_include.keys():
                            assets_to_include[group].append(asset)
                        else:
                            assets_to_include[group] = [asset]
                    asset_variable_map[asset] = type_variable_map[asset_type]

        assets_list = [asset for a_type in assets_to_include.values() for asset in a_type]
        unused_asset = [
            asset.asset_type
            for asset in self.esdl_assets.values()
            if asset.asset_type not in assets_without_control
            if asset.name not in assets_list
            if asset.name not in available_timeseries
        ]
        assert (
            len(unused_asset) == 0
        ), f"Asset types: {unused_asset} are not included in controls of the simulator"

        asset_info = {
            "assets_to_include": assets_to_include,
            "assets_to_include_list": assets_list,
            "asset_variable_map": asset_variable_map,
        }
        return asset_info

    def __create_merit_path_goals(self, asset_info, max_value_merit, index_start_of_priority):
        """
        This method creates the goals for every asset that is based on the marginal cost and the
        relevant variable for that asset. Depending on the type of asset, the goal is a minimisation
        or maximisation.
        :param asset_info:
        :param max_value_merit:
        :param index_start_of_priority:
        :return:
        """
        goals = []
        assets_to_include = asset_info["assets_to_include"]
        assets_list = asset_info["assets_to_include_list"]
        asset_merit = asset_info["asset_merit"]
        asset_variable_map = asset_info["asset_variable_map"]
        for asset in assets_list:
            index_s = asset_merit["asset_name"].index(f"{asset}")
            marginal_priority = (
                index_start_of_priority + max_value_merit - asset_merit["merit_order"][index_s]
            )
            assert (
                marginal_priority >= index_start_of_priority
            ), "Priorities assigned must be smaller than the total number of producers"

            if asset in [
                *assets_to_include.get("source", []),
            ]:
                variable_name = f"{asset}.{asset_variable_map[asset]}"

                goals.append(
                    MinimizeSourcesGoalMerit(
                        variable_name,
                        marginal_priority,
                        self.bounds()[variable_name],
                        self.variable_nominal(variable_name),
                    )
                )
            elif asset in assets_to_include.get("conversion", []):
                variable_name = f"{asset}.{asset_variable_map[asset]}"
                index_s = asset_merit["asset_name"].index(f"{asset}_prod")
                marginal_priority_source = (
                    index_start_of_priority + max_value_merit - asset_merit["merit_order"][index_s]
                )
                goals.append(
                    MinimizeSourcesGoalMerit(
                        variable_name,
                        marginal_priority_source,
                        self.bounds()[variable_name],
                        self.variable_nominal(variable_name),
                    )
                )
                variable_name = f"{asset}.Gas_mass_flow_out"
                goals.append(
                    MaximizeDemandGoalMerit(
                        variable_name,
                        marginal_priority,
                        self.bounds()[variable_name],
                        self.variable_nominal(variable_name),
                    )
                )
            elif asset in assets_to_include.get("demand", []):
                variable_name = f"{asset}.{asset_variable_map[asset]}"

                goals.append(
                    MaximizeDemandGoalMerit(
                        variable_name,
                        marginal_priority,
                        self.bounds()[variable_name],
                        self.variable_nominal(variable_name),
                    )
                )
            elif asset in assets_to_include.get("storage", []):
                # TODO: should use separate variable for charging and discharging

                # charging acts as consumer
                # Marginal costs for discharging > marginal cost for charging
                variable_name = f"{asset}.{asset_variable_map[asset]['charge']}"

                func_range = self.bounds()[variable_name]
                v1 = _extract_values_timeseries(func_range[0], "min")
                v2 = _extract_values_timeseries(func_range[1], "max")
                func_range = (v1, v2)

                goals.append(
                    MaximizeStorageGoalMerit(
                        variable_name,
                        marginal_priority,
                        func_range,
                        self.variable_nominal(variable_name),
                    )
                )

                # discharging acts as producer
                # Marginal costs for discharging should be larger than marginal cost for charging
                # TODO: add check on the marginal costs for charging/discharging
                index_s = asset_merit["asset_name"].index(f"{asset}_discharge")
                marginal_priority = (
                    index_start_of_priority + max_value_merit - asset_merit["merit_order"][index_s]
                )
                assert (
                    marginal_priority >= index_start_of_priority
                ), "Priorities assigned must be smaller than the total number of producers"

                variable_name = f"{asset}{asset_variable_map[asset]['discharge']}"

                func_range = self.bounds()[variable_name]
                v1 = _extract_values_timeseries(func_range[0], "min")
                v2 = _extract_values_timeseries(func_range[1], "max")
                func_range = (v1, v2)

                goals.append(
                    MinimizeStorageGoalMerit(
                        variable_name,
                        marginal_priority,
                        func_range,
                        self.variable_nominal(variable_name),
                    )
                )
            else:
                raise Exception(
                    f"No goal was set for {asset}, while a priority was provided. This asset group "
                    f"still has to be added to the goals."
                )
        return goals

    def __merit_path_goals(self):
        """
        This method organizes the goals and assigns their priorities
        The first two priorities are reserved for matching of demand. Thereby the priorities of the
         other goals to maximize specific producers and minimize demand, start at priority 3.
        :return:
        """

        # TODO: improve the asset_types_to_include and esdl_assets_to_include
        asset_types_to_include = {
            "source": ["electricity_source", "gas_source"],
            "demand": ["electricity_demand", "gas_demand"],
            "conversion": ["electrolyzer"],
            "storage": ["gas_tank_storage", "electricity_storage"],
        }

        assets_without_control = ["Pipe", "ElectricityCable", "Joint", "Bus"]

        # TODO also include other assets than producers, e.g. storage, conversion and possible
        #  demand for the ones without a profile
        # TODO exclude producers from merit order if they have a profile, even if a marginal cost
        #  is set

        # Storage: charge priority should be higher than discharge priority
        asset_info = self.__create_asset_list_controls(
            asset_types_to_include, assets_without_control
        )

        asset_info["asset_merit"] = self.__merit_controls(asset_info["assets_to_include_list"])
        max_value_merit = max(asset_info["asset_merit"]["merit_order"])

        # Priority 1 & 2 reserved for target demand goal & additional goal like matching
        # producer profile (without merit order)
        index_start_of_priority = 3

        goals = self.__create_merit_path_goals(asset_info, max_value_merit, index_start_of_priority)

        return goals

    def path_goals(self):
        goals = super().path_goals().copy()

        goals.extend(self.__merit_path_goals())

        return goals

    def energy_system_options(self):
        options = super().energy_system_options()

        self.gas_network_settings["head_loss_option"] = HeadLossOption.LINEARIZED_N_LINES_EQUALITY
        self.gas_network_settings["network_type"] = NetworkSettings.NETWORK_TYPE_HYDROGEN
        self.gas_network_settings["minimize_head_losses"] = False
        self.gas_network_settings["maximum_velocity"] = 60.0
        options["include_asset_is_switched_on"] = True
        options["estimated_velocity"] = 7.5

        options["gas_storage_discharge_variables"] = True

        return options

    def path_constraints(self, ensemble_member):
        """
        Constraints to limit producer production in case timeseries for the production exist,
        relevant when the first goals is not to match profile.
        """
        constraints = super().path_constraints(ensemble_member)

        return constraints

    def constraints(self, ensemble_member):
        """
        Add equality constraints to enforce a cyclic energy balance [J] between the end and the
        start of the time horizon used as well an inequality constraint to enforce no milp supply
        [W] to the netwok in the 1st time step
        """
        constraints = super().constraints(ensemble_member)

        # TODO: cyclic constraints

        return constraints

    def __merit_controls(self, assets_list):
        attributes = {
            "asset_name": [],
            "merit_order": [],
        }

        assets = self.esdl_assets
        for a in assets.values():
            if a.name in assets_list:
                attributes["asset_name"].append(a.name)
                try:
                    attributes["merit_order"].append(
                        a.attributes["costInformation"].marginalCosts.value
                    )
                    if (
                        a.asset_type == "Electrolyzer"
                    ):  # electrolyzer would require minimisation of electricity right after
                        # maximisation gas, therefore a merit_order with a very small adjusted
                        # value is used (e.g. 1e-6) to ensure these goals are right after each
                        # other and no other goals are in between.
                        attributes["asset_name"].append(f"{a.name}_prod")
                        attributes["merit_order"].append(
                            a.attributes["costInformation"].marginalCosts.value - 1e-6
                        )

                except AttributeError:
                    try:
                        attributes["merit_order"].append(
                            a.attributes["controlStrategy"].marginalChargeCosts.value
                        )
                        attributes["merit_order"].append(
                            a.attributes["controlStrategy"].marginalDischargeCosts.value
                        )
                        attributes["asset_name"].append(f"{a.name}_discharge")
                    except AttributeError:
                        raise Exception(f"Asset: {a.name} does not have a merit order specified")

                last_merit_order = attributes["merit_order"][-1]

                if attributes["merit_order"][-1] <= 0.0:
                    raise Exception(
                        "The specified producer usage merit order must be a "
                        f"positve integer value, producer name:{a.name}, current "
                        f"specified merit value: {last_merit_order}"
                    )

        # update merit order to be integers and have maximumally one step in priority
        set_merit = set(attributes["merit_order"])
        order_prio = list(set_merit)
        order_prio.sort()
        map_prio = dict(zip(order_prio, list(range(1, len(set_merit) + 1))))
        a = [map_prio[v] for v in attributes["merit_order"]]
        attributes["merit_order"] = a
        return attributes

    def solver_options(self):
        options = super().solver_options()
        options["casadi_solver"] = self._qpsol

        options["solver"] = "highs"
        highs_options = options["highs"] = {}
        highs_options["presolve"] = "on"

        return options

    def priority_started(self, priority):
        goals_print = set()
        for goal in [*self.path_goals(), *self.goals()]:
            if goal.priority == priority:
                goals_print.update([str(type(goal))])
        logger.info(f"{goals_print}")
        self.__priority = priority
        self.__priority_timer = time.time()

        super().priority_started(priority)

    def priority_completed(self, priority):
        super().priority_completed(priority)

        self._hot_start = True

        time_taken = time.time() - self.__priority_timer
        self._priorities_output.append(
            (
                priority,
                time_taken,
                True,
                self.objective_value,
                self.solver_stats,
            )
        )
        logger.info(f"Goal with priority {priority} has been completed")
        if priority == 1 and self.objective_value > 1e-6:
            raise RuntimeError(
                f"The heating demand is not matched, objective value is {self.objective_value}"
            )

    def __state_vector_scaled(self, variable, ensemble_member):
        canonical, sign = self.alias_relation.canonical_signed(variable)
        return (
            self.state_vector(canonical, ensemble_member) * self.variable_nominal(canonical) * sign
        )

    # TODO: post will be created later
    # def post(self):
    #     super().post()
    #     self._write_updated_esdl(self.get_energy_system_copy(), optimizer_sim=True)


# -------------------------------------------------------------------------------------------------
class MultiCommoditySimulatorHIGHS(MultiCommoditySimulator):

    def solver_options(self):
        options = super().solver_options()
        options["solver"] = "highs"

        return options


class MultiCommoditySimulatorNoLosses(MultiCommoditySimulator):
    def energy_system_options(self):
        options = super().energy_system_options()

        self.gas_network_settings["head_loss_option"] = HeadLossOption.NO_HEADLOSS
        self.gas_network_settings["minimize_head_losses"] = False
        options["include_electric_cable_power_loss"] = False

        return options

    def solver_options(self):
        # For some cases the presolve of the HIGHS solver makes this problem infeasible, therefore
        # the presolve is turned off.
        options = super().solver_options()
        options["solver"] = "highs"
        highs_options = options["highs"] = {}
        highs_options["presolve"] = "off"

        return options


def run_sequatially_staged_simulation(
    multi_commodity_simulator_class, simulation_window_size=2, *args, **kwargs
):
    """
    This function is to run the MultiCommoditySimulator class in a staged manner where the stages
    are sequantial parts of the time horizon. This is done to make the problem smaller and allow
    for quicker run of the optimization/simulation. To ensure a physically sound result the
    variables that affect the outcome of the next stage are constrained by setting bounds, e.g. the
    amount of stored energy in a storage.

    Parameters
    ----------
    multi_commodity_simulator_class : The problem class to run
    simulation_window_size : The amount of indices to run in a single stage

    Returns
    -------
    The results dict where all the time results are concatenated.
    """

    class MultiCommoditySimulatorTimeSequential(multi_commodity_simulator_class):
        """
        This Problem class is used to run the MultiCommoditySimulator class in a sequantial manner
        to reduce computational time. This class enables this by allowing to run a part of the
        timeseries and setting bounds on the (initial-)state variables.
        """

        def __init__(self, *args, **kwargs):
            super().__init__(*args, **kwargs)

            self.__start_time_index = kwargs.get("start_index", None)
            self.__end_time_index = kwargs.get("end_index", None)
            self._full_time_series = None

            self.__storage_initial_state_bounds = kwargs.get("storage_initial_state_bounds", {})

        def times(self, variable=None) -> np.ndarray:
            """
            In this function the part of the time-horizon is enforced. Note that the full
            time-horizon is also set to an internal member for enabling setting bounds to the next
            sequantial optimization.
            """
            if self._full_time_series is None:
                self._full_time_series = super().times(variable)

            if self.__start_time_index is not None and self.__end_time_index is not None:
                return super().times(variable)[self.__start_time_index : self.__end_time_index]
            elif self.__start_time_index is not None:
                return super().times(variable)[self.__start_time_index :]
            elif self.__end_time_index is not None:
                return super().times(variable)[: self.__end_time_index]
            else:
                return super().times(variable)

        def bounds(self):
            """
            Here we set bounds on the initial state to ensure that the sequantial simulation is a
            physcially valid.
            """
            bounds = super().bounds()
            bounds.update(self.__storage_initial_state_bounds)
            return bounds

    # Note that the window size should be larger than 1 otherwise this function is has no
    # benefit and the writing of the results dict will fail.
    assert simulation_window_size >= 2

    total_results = None
    # This is an initial value for end_time, will be corrected after the first stage
    end_time = 2 * simulation_window_size
    end_time_confirmed = False
    storage_initial_state_bounds = {}

    # TODO: make this dict complete for all relevant assets and their associated variables.
    constrained_assets = {
        "electricity_storage": ["Stored_electricity", "Effective_power_charging"],
        "gas_tank_storage": ["Stored_gas_mass", "Gas_tank_flow"],
    }

    tic = time.time()
    for simulated_window in range(0, end_time, simulation_window_size):

        # Note that the end time is not necessarily a multiple of simulation_window_size
        sub_end_time = min(end_time, simulated_window + simulation_window_size)

        # max operation for start_index to avoid the overlap function in the first stage
        solution = run_optimization_problem(
            MultiCommoditySimulatorTimeSequential,
            start_index=max(simulated_window - 1, 0),
            end_index=sub_end_time,
            storage_initial_state_bounds=storage_initial_state_bounds,
            **kwargs,
        )
        if not end_time_confirmed:
            end_time = len(solution._full_time_series)
            end_time_confirmed = True
        results = solution.extract_results()

        # TODO: check if we now capture all relevant variables.
        if total_results is None:
            total_results = results
            aliases = solution.alias_relation._canonical_variables_map
            bounds = solution.bounds()
            parameters = solution.parameters(0)
        else:
            for key, data in results.items():
                if len(total_results[key]) > 1:
                    total_results[key] = np.concatenate((total_results[key], data[1:]))

        if sub_end_time < end_time:
            for asset_type, variables in constrained_assets.items():
                for asset in solution.energy_system_components.get(asset_type, []):
                    sub_time_series = solution._full_time_series[
                        simulated_window
                        - 1
                        + simulation_window_size : min(
                            end_time, simulated_window + 2 * simulation_window_size
                        )
                    ]
                    for variable in variables:
                        lb_value = _extract_values_timeseries(
                            solution.bounds()[f"{asset}.{variable}"][0], "min"
                        )
                        ub_value = _extract_values_timeseries(
                            solution.bounds()[f"{asset}.{variable}"][1], "max"
                        )
                        lb_values = [lb_value] * len(sub_time_series)
                        ub_values = [ub_value] * len(sub_time_series)
                        lb_values[0] = ub_values[0] = results[f"{asset}.{variable}"][-1]
                        lb = Timeseries(sub_time_series, lb_values)
                        ub = Timeseries(sub_time_series, ub_values)
                        storage_initial_state_bounds[f"{asset}.{variable}"] = (lb, ub)

    print(time.time() - tic)

    return OptimisationOverview(total_results, bounds, parameters, aliases)


# -------------------------------------------------------------------------------------------------
@main_decorator
def main(runinfo_path, log_level):
    logger.info("Run Network Simulator")

    kwargs = {
        "write_result_db_profiles": False,
        "database_connections": [
            {
                "access_type": DBAccesType.WRITE,
                "influxdb_host": "localhost",
                "influxdb_port": 8086,
                "influxdb_username": None,
                "influxdb_password": None,
                "influxdb_ssl": False,
                "influxdb_verify_ssl": False,
            },
        ],
    }

    _ = run_optimization_problem(
        MultiCommoditySimulatorHIGHS,
        esdl_run_info_path=runinfo_path,
        log_level=log_level,
        **kwargs,
    )


if __name__ == "__main__":
    # main()
    import tests.models.emerge.src.example as example

    base_folder = Path(example.__file__).resolve().parent.parent
    solution = run_optimization_problem(
        MultiCommoditySimulatorNoLosses,
        base_folder=base_folder,
        esdl_file_name="emerge_priorities_withoutstorage.esdl",
        esdl_parser=ESDLFileParser,
        profile_reader=ProfileReaderFromFile,
        input_timeseries_file="timeseries.csv",
    )



from pathlib import Path
from unittest import TestCase

from mesido.esdl.esdl_parser import ESDLFileParser
from mesido.esdl.profile_parser import ProfileReaderFromFile
from mesido.util import run_esdl_mesido_optimization

import numpy as np

from utils_tests import demand_matching_test, electric_power_conservation_test, feasibility_test


class TestMILPElectricSourceSinkStorage(TestCase):
    def test_source_sink_storage(self):
        """
        Tests for an electricity network that consist out of a source, a sink and storage,
        connected with electricity cables.

        Checks:
        - Check that the efficiency for charging and discharging is used proper to determine the
        consumed and stored power
        - Check that the battery is used to match demand
        - Check that charging and discharging properly connected to the network (direction)
        - Check that the is_charging variable is set correctly
        - Checks that the use of is_charging variable gives the same results as without the use
        of the binary is_charging variable.

        """

        import models.unit_cases_electricity.battery.src.example as example
        from models.unit_cases_electricity.battery.src.example import ElectricityProblem

        base_folder = Path(example.__file__).resolve().parent.parent
        tol = 1e-10

        solution = run_esdl_mesido_optimization(
            ElectricityProblem,
            base_folder=base_folder,
            esdl_file_name="source_sink_storage.esdl",
            esdl_parser=ESDLFileParser,
            profile_reader=ProfileReaderFromFile,
            input_timeseries_file="timeseries.csv",
        )

        results = solution.extract_results()
        parameters = solution.parameters(0)

        feasibility_test(solution)

        demand_matching_test(solution, results)
        electric_power_conservation_test(solution, results)

        class ElectricityProblemDisc(ElectricityProblem):
            def energy_system_options(self):
                options = super().energy_system_options().copy()
                options["electricity_storage_discrete_charge_variables"] = True
                return options

        solution_disc = run_esdl_mesido_optimization(
            ElectricityProblemDisc,
            base_folder=base_folder,
            esdl_file_name="source_sink_storage.esdl",
            esdl_parser=ESDLFileParser,
            profile_reader=ProfileReaderFromFile,
            input_timeseries_file="timeseries.csv",
        )

        results_disc = solution_disc.extract_results()

        storage_name = solution.energy_system_components.get("electricity_storage")[0]
        charge_eff = parameters[f"{storage_name}.charge_efficiency"]
        discharge_eff = parameters[f"{storage_name}.discharge_efficiency"]
        eff_power_change_bat = results[f"{storage_name}.Effective_power_charging"]
        power_bat_network = results[f"{storage_name}.ElectricityIn.Power"]
        power_discharging = results[f"{storage_name}.Power_discharging"]
        power_charging = results[f"{storage_name}.Power_charging"]
        stored_el = results[f"{storage_name}.Stored_electricity"]

        power_cable_bat = results["ElectricityCable_91c1.ElectricityOut.Power"]
        np.testing.assert_allclose(power_cable_bat, power_bat_network, atol=tol)
        np.testing.assert_allclose(power_bat_network, power_charging - power_discharging, atol=tol)

        power_discharging_disc = results_disc[f"{storage_name}.Power_discharging"]
        power_charging_disc = results_disc[f"{storage_name}.Power_charging"]
        np.testing.assert_allclose(power_discharging[1:], power_discharging_disc[1:], atol=tol)
        np.testing.assert_allclose(power_charging[1:], power_charging_disc[1:], atol=tol)

        for sol in [solution, solution_disc]:
            count_discrete_path, count_discrete_var = 0.0, 0.0
            for var in sol.path_variables:
                if sol.variable_is_discrete(var.name()):
                    count_discrete_path += 1
            for var in sol.extra_variables:
                if sol.variable_is_discrete(var.name()):
                    count_discrete_var += 1
            np.testing.assert_allclose(count_discrete_var, 7.0)
            if sol == solution:
                np.testing.assert_allclose(0.0, count_discrete_path, atol=tol)
            else:
                np.testing.assert_allclose(1.0, count_discrete_path, atol=tol)

        is_charging = np.asarray([float(i > 0) for i in power_charging])
        # if battery is charging (1), ElectricityIn.Power and effective_power charging should be
        # positive, else negative
        bigger_then = all(is_charging * eff_power_change_bat >= 0)
        smaller_then = all((1 - is_charging) * eff_power_change_bat <= 0)
        self.assertTrue(bigger_then)
        self.assertTrue(smaller_then)

        bigger_then = all(is_charging * power_bat_network >= 0)
        smaller_then = all((1 - is_charging) * power_bat_network <= 0)
        self.assertTrue(bigger_then)
        self.assertTrue(smaller_then)

        # battery should be charging at atleast one timestep to overcome the difference between max
        # production and demand
        self.assertTrue(sum(is_charging) >= 1)

        # stored electricity starts at 0
        np.testing.assert_allclose(
            stored_el[0],
            0.0,
            err_msg="The battery should be empty at the start to check if it operates and "
            "predicted",
        )

        # stored electricity change should be equal to the effective power change
        stored_change = stored_el[1:] - stored_el[:-1]
        np.testing.assert_allclose(eff_power_change_bat[1:] * 3600, stored_change, atol=tol)

        # effective power change while charging should be equal to efficiency * powerIn
        np.testing.assert_allclose(
            eff_power_change_bat * is_charging,
            power_bat_network * is_charging * charge_eff,
            atol=tol,
        )
        # effective power change while discharging should be equal to powerOut/efficiency
        np.testing.assert_allclose(
            eff_power_change_bat * (1 - is_charging),
            power_bat_network * (1 - is_charging) / discharge_eff,
            atol=tol,
        )



from mesido.pycml import Variable
from mesido.pycml.pycml_mixin import add_variables_documentation_automatically

from numpy import nan

from .electricity_base import ElectricityTwoPort
from .._internal import BaseAsset


@add_variables_documentation_automatically
class ElectricityCable(ElectricityTwoPort, BaseAsset):
    """
    The electricity cable component is used to model voltage and power drops in the electricity
    lines. We model the power losses by over estimating them with the maximum current. We ensure
    that the power is always less than what the current is able to carry by an equality constraint
    at the demand where we enforce the minimum voltage.

    Variables created:
        {add_variable_names_for_documentation_here}

    Parameters:
        name : The name of the asset. \n
        modifiers : Dictionary with asset information.
    """

    def __init__(self, name, **modifiers):
        super().__init__(name, **modifiers)

        self.component_type = "electricity_cable"
        self.disconnectable = False

        self.length = 1.0

        # Powerloss with inequality in the milp-mixin
        # values for NAYY 4x50 SE
        # from: https://pandapower.readthedocs.io/en/v2.6.0/std_types/basic.html
        self.max_current = nan
        self.min_voltage = nan
        self.max_voltage = self.min_voltage * 2.0
        self.nominal_current = nan
        self.nominal_voltage = nan
        self.r = nan
        self.nominal_voltage_loss = (self.nominal_current * self.r * self.nominal_voltage) ** 0.5
        self.power_loss_nominal = self.r * self.max_current * self.nominal_current
        # We accept lower accuracy in the loss computation to improve scaling in case the nominals
        # are very far apart. Typically, when a short cable has a very high max capacity.
        if self.power_loss_nominal / self.ElectricityIn.Power.nominal < 1.0e-4:
            self.power_loss_nominal = self.ElectricityIn.Power.nominal * 1.0e-4
        self.add_variable(Variable, "Power_loss", min=0.0, nominal=self.power_loss_nominal)
        self.add_variable(Variable, "I", nominal=self.nominal_current)
        self.add_variable(Variable, "V_loss", nominal=self.nominal_voltage_loss)

        # TODO: if one wants to include the option for cable_voltage_losses to be false as with
        #  cable_power_losses, then the next equation and V_loss variable instantiation should move
        #  to electricity_physics_mixin, to ensure proper scaling.
        self.add_equation(
            (
                (self.ElectricityOut.V - (self.ElectricityIn.V - self.V_loss))
                / self.nominal_voltage_loss
            )
        )
        self.add_equation(((self.ElectricityIn.I - self.ElectricityOut.I) / self.nominal_current))
        self.add_equation(((self.ElectricityIn.I - self.I) / self.nominal_current))
        self.add_equation(
            (
                (self.ElectricityOut.Power - (self.ElectricityIn.Power - self.Power_loss))
                / (self.ElectricityIn.Power.nominal * self.Power_loss.nominal) ** 0.5
            )
        )



from mesido.pycml.pycml_mixin import add_variables_documentation_automatically

from ._non_storage_component import _NonStorageComponent


@add_variables_documentation_automatically
class _NonStorageComponentSinkType(_NonStorageComponent):
    """
    A non storage component is an asset type that does not have states that have interdependent
    time-steps, it does have time-varying states like Heat_flow. In the
    NonStorageSinkTypeComponent we add a limit on the headloss.

    Variables created:
        {add_variable_names_for_documentation_here}

    Parameters:
        name : The name of the asset. \n
        modifiers : Dictionary with asset information.
    """

    def __init__(self, name, **modifiers):
        super().__init__(name, **modifiers)

        self.dH.max = 0.0



from typing import Dict, TYPE_CHECKING, Tuple, Type

if TYPE_CHECKING:  # This prevents circular referencing between this file and potential_errors.py
    from mesido.potential_errors import ErrorMessage, MesidoAssetIssueType


class MesidoAssetIssueError(Exception):
    """
    This class is used as a MESIDO defined exception class.
    """

    error_type: "MesidoAssetIssueType"
    general_issue: str
    message_per_asset_id: Dict[str, "ErrorMessage"]

    def __init__(
        self,
        general_issue: str,
        error_type: "MesidoAssetIssueType",
        message_per_asset_id: Dict[str, "ErrorMessage"],
    ):

        self.error_type = error_type
        self.general_issue = general_issue
        self.message_per_asset_id = message_per_asset_id

    def __str__(self) -> str:
        """
        Returns
        -------
        A string containing the following:
        - Error type (e.g.HEAT_DEMAND_POWER)
        - General meassage about the group of error messages per asset id
        - Asset id
        - Error message for the applicable asset
        """
        return_string = f"{self.error_type}: {self.general_issue}"
        for keys, values in self.message_per_asset_id.items():
            return_string += f"\nAsset id: {keys}"
            return_string += f" {values}"

        return return_string

    # Currently this function has been added for Design Toolkit back end purposes
    def __reduce__(
        self,
    ) -> Tuple[
        Type["MesidoAssetIssueError"], Tuple[str, "MesidoAssetIssueType", Dict[str, "ErrorMessage"]]
    ]:
        """Reduce an instance of this class to pickable components.

        This is required so the class can be serialized and deserialized in the pickle format.

        Returns
        -------
        A tuple containing:
        - The class
        - Separate arguments of this instance
        """
        return (
            MesidoAssetIssueError,
            (self.general_issue, self.error_type, self.message_per_asset_id),
        )



from mesido.pycml import Variable
from mesido.pycml.component_library.milp._internal import HeatComponent
from mesido.pycml.pycml_mixin import add_variables_documentation_automatically

from numpy import nan

from .heat_port import HeatPort


@add_variables_documentation_automatically
class HeatTwoPort(HeatComponent):
    """
    The HeatTwoPort component is used as a base for interaction with one hydraulically coupled
    system. As heat networks are closed systems we always need two ports to model both the in and
    out going flow in the system.

    Variables created:
        {add_variable_names_for_documentation_here}

    Parameters:
        name : The name of the asset. \n
        modifiers : Dictionary with asset information.
    """

    def __init__(self, name, **modifiers):
        super().__init__(name, **modifiers)

        self.Q_nominal = 1.0
        self.T_supply = nan
        self.T_return = nan
        self.T_supply_id = -1
        self.T_return_id = -1
        self.dT = self.T_supply - self.T_return
        self.cp = 4200.0
        self.rho = 988.0
        self.nominal_pressure = 16.0e5

        self.add_variable(HeatPort, "HeatIn")
        self.add_variable(HeatPort, "HeatOut")

        self.add_variable(Variable, "Q", nominal=self.Q_nominal)

        self.add_equation(self.HeatIn.Q - self.Q)
        self.add_equation(self.HeatIn.Q - self.HeatOut.Q)

        self.add_variable(Variable, "dH")
        self.add_equation(self.dH - (self.HeatOut.H - self.HeatIn.H))



import math
from pathlib import Path
from unittest import TestCase

from mesido.esdl.esdl_parser import ESDLFileParser
from mesido.esdl.profile_parser import ProfileReaderFromFile
from mesido.util import run_esdl_mesido_optimization

import numpy as np

from utils_tests import demand_matching_test, energy_conservation_test, heat_to_discharge_test


class TestGasBoiler(TestCase):

    def asset_cost_calculation_tests(self, solution, results, parameters):
        # Check the cost components of GasHeater
        for asset in [
            *solution.energy_system_components.get("heat_source", []),
        ]:
            esdl_asset = solution.esdl_assets[solution.esdl_asset_name_to_id_map[f"{asset}"]]
            costs_esdl_asset = esdl_asset.attributes["costInformation"]

            # investment cost
            investment_cost_info = costs_esdl_asset.investmentCosts.value
            investment_cost = investment_cost_info * results[f"{asset}__max_size"] / 1.0e6
            np.testing.assert_allclose(
                investment_cost, results[f"{asset}__investment_cost"], atol=1.0e-7
            )

            # installation cost
            installation_cost = costs_esdl_asset.installationCosts.value
            np.testing.assert_allclose(installation_cost, results[f"{asset}__installation_cost"])

            # variable operational cost
            timesteps_hr = np.diff(solution.times())
            variable_operational_cost = 0.0
            var_op_costs = costs_esdl_asset.variableOperationalCosts.value
            for ii in range(1, len(solution.times())):
                variable_operational_cost += (
                    var_op_costs
                    * results[f"{asset}.Gas_demand_mass_flow"][ii]
                    / parameters[f"{asset}.density_normal"]
                    * timesteps_hr[ii - 1]
                )
            np.testing.assert_allclose(
                variable_operational_cost, results[f"{asset}__variable_operational_cost"]
            )

    def test_gas_heat_source_gas(self):
        """
        This tests checks the gas boiler for the standard checks and the energy conservation over
        the commodity change.

        Checks:
        1. demand is matched
        2. energy conservation in the network
        3. heat to discharge
        4. energy conservation over the commodity
        5. GasHeater cost components

        """
        import models.source_pipe_sink.src.double_pipe_heat as example
        from models.source_pipe_sink.src.double_pipe_heat import SourcePipeSink

        base_folder = Path(example.__file__).resolve().parent.parent

        heat_problem = run_esdl_mesido_optimization(
            SourcePipeSink,
            base_folder=base_folder,
            esdl_file_name="sourcesink_with_gasboiler.esdl",
            esdl_parser=ESDLFileParser,
            profile_reader=ProfileReaderFromFile,
            input_timeseries_file="timeseries_import.csv",
        )
        results = heat_problem.extract_results()
        parameters = heat_problem.parameters(0)
        bounds = heat_problem.bounds()

        demand_matching_test(heat_problem, results)
        energy_conservation_test(heat_problem, results)
        heat_to_discharge_test(heat_problem, results)

        # check energy conservation over the commodity
        np.testing.assert_array_less(0.0, results["GasHeater_f713.Heat_source"])
        np.testing.assert_array_less(0.0, results["GasProducer_82ec.Gas_source_mass_flow"])
        np.testing.assert_array_less(
            parameters["GasHeater_f713.energy_content"]
            * results["GasHeater_f713.GasIn.mass_flow"]
            * parameters["GasHeater_f713.efficiency"]
            / 1000.0,  # [J/kg] * [g/s] / 1000.0 = [J/s]
            results["GasHeater_f713.Heat_source"] + 1.0e-6,
        )

        # check if the maximum gas velocity set in problem is used to determine bounds on pipes
        v_max_gas = heat_problem.gas_network_settings[
            "maximum_velocity"
        ]  # m/s maximum velocity set in problem.
        np.testing.assert_allclose(
            bounds["Pipe_a7b5.GasIn.Q"][1],
            parameters["Pipe_a7b5.diameter"] ** 2 / 4 * math.pi * v_max_gas,
        )

        # check the cost components of GasHeater
        self.asset_cost_calculation_tests(heat_problem, results, parameters)

    def test_heat_source_gas(self):
        """
        This tests checks the gas boiler without gas inlet port for the standard checks
        and GasHeater cost components

        Checks:
        1. demand is matched
        2. energy conservation in the network
        3. heat to discharge
        4. energy conservation over the commodity
        5. GasHeater cost components

        """
        import models.source_pipe_sink.src.double_pipe_heat as example
        from models.source_pipe_sink.src.double_pipe_heat import SourcePipeSink

        base_folder = Path(example.__file__).resolve().parent.parent

        solution = run_esdl_mesido_optimization(
            SourcePipeSink,
            base_folder=base_folder,
            esdl_file_name="sourcesink_with_gasboiler_no_gas.esdl",
            esdl_parser=ESDLFileParser,
            profile_reader=ProfileReaderFromFile,
            input_timeseries_file="timeseries_import.csv",
        )
        results = solution.extract_results()
        parameters = solution.parameters(0)

        demand_matching_test(solution, results)
        energy_conservation_test(solution, results)
        heat_to_discharge_test(solution, results)

        # check energy conservation over the commodity
        np.testing.assert_array_less(0.0, results["GasHeater_f713.Heat_source"])
        np.testing.assert_array_less(0.0, results["GasHeater_f713.Gas_demand_mass_flow"])
        np.testing.assert_array_less(
            parameters["GasHeater_f713.energy_content"]
            * results["GasHeater_f713.Gas_demand_mass_flow"]
            * parameters["GasHeater_f713.efficiency"]
            / 1000.0,  # [J/kg] * [g/s] / 1000.0 = [J/s]
            results["GasHeater_f713.Heat_source"] + 1.0e-6,
        )

        # Check the cost components of GasHeater
        self.asset_cost_calculation_tests(solution, results, parameters)


if __name__ == "__main__":

    a = TestGasBoiler()
    a.test_gas_heat_source_gas()
    a.test_heat_source_gas()



from pathlib import Path
from unittest import TestCase

import esdl

import mesido._darcy_weisbach as darcy_weisbach
from mesido.esdl.esdl_parser import ESDLFileParser
from mesido.esdl.profile_parser import ProfileReaderFromFile
from mesido.workflows import (
    EndScenarioSizing,
    EndScenarioSizingDiscounted,
    EndScenarioSizingStaged,
    run_end_scenario_sizing,
)
from mesido.workflows.grow_workflow import EndScenarioSizingHeadLossStaged
from mesido.workflows.utils.error_types import NetworkErrors

import numpy as np

from rtctools.util import run_optimization_problem

from utils_tests import demand_matching_test


class TestEndScenarioSizing(TestCase):

    def test_heat_exchanger_sizing(self):
        """
        Check heat exchanger can be sized in EndScenarioSizingStaged problem.
        After optimization asset state and capacity attributes are changed.

        Checks:
        - max_size variable of the asset is calculated
        - heat exchanger state attribute is changed
        - heat exchanger capacity attribute is updated
        """
        import models.heat_exchange.src.run_heat_exchanger as run_heat_exchanger

        base_folder = Path(run_heat_exchanger.__file__).resolve().parent.parent

        solution = run_end_scenario_sizing(
            EndScenarioSizingStaged,
            base_folder=base_folder,
            esdl_file_name="heat_exchanger_with_costs.esdl",
            esdl_parser=ESDLFileParser,
        )
        results = solution.extract_results()

        # Check heat exchanger is sized
        np.testing.assert_allclose(
            max(results["HeatExchange_39ed.Secondary_heat"]), results["HeatExchange_39ed__max_size"]
        )

        # Check heat exchanger state attribute is changed from OPTIONAL
        # to ENABLED after the optimization
        energy_system = solution._ESDLMixin__energy_system_handler.energy_system
        asset = solution._name_to_asset(energy_system, "HeatExchange_39ed")
        np.testing.assert_equal(esdl.AssetStateEnum.ENABLED, asset.state)

        # Check heat exchanger capacity attribute is updated
        # with max_size variable after the optimization
        np.testing.assert_allclose(results["HeatExchange_39ed__max_size"], asset.capacity)

    @classmethod
    def setUpClass(cls) -> None:
        import models.test_case_small_network_ates_buffer_optional_assets.src.run_ates as run_ates

        base_folder = Path(run_ates.__file__).resolve().parent.parent

        # This is an optimization done over a full year with timesteps of 5 days and hour timesteps
        # for the peak day
        cls.solution = run_optimization_problem(
            EndScenarioSizing,
            base_folder=base_folder,
            esdl_file_name="test_case_small_network_with_ates_with_buffer_all_optional.esdl",
            esdl_parser=ESDLFileParser,
            profile_reader=ProfileReaderFromFile,
            input_timeseries_file="Warmte_test.csv",
        )
        cls.results = cls.solution.extract_results()

    def test_end_scenario_sizing(self):
        """
        Check if the EndScenarioSizingHIGHS sizing workflow is behaving as expected. This is an
        optimization done over a full year with timesteps of 5 days and hour timesteps for the peak
        day.

        Checks:
        - demand matching
        - that the available pipe classes were adapted
        - minimum velocity setting
        - Cyclic behaviour for ATES
        - That buffer tank is only used on peak day
        - Check if TCO goal included the desired cost components.


        Missing:
        - Link ATES t0 utilization to state of charge at end of year for optimizations over one
        year.
        """

        # In the future we want to check the following
        # Is the timeline correctly converted, correct peak day, correct amount of timesteps, etc.
        # Check whether expected assets are disabled
        # Check the optimal size of assets
        # Check the cost breakdown, check whether all the enabled assets are in the cost breakdown
        # Check that computation time is within expected bounds

        # Check whehter the heat demand is matched
        demand_matching_test(self.solution, self.results)

        # Check that indeed the available pipe classes were adapted based on expected flow
        # Pipe connected to a demand
        assert self.solution.pipe_classes("Pipe2")[0].name == "DN150"  # initially DN->None
        assert self.solution.pipe_classes("Pipe2")[-1].name == "DN250"  # initially DN450
        # Check that the available pipe classes are also limited for pipes in the return network
        # that do not have the related attribute assigned and are therefore not in the set of
        # self.cold_pipes.
        assert len(self.solution.unrelated_pipes) >= 1.0
        assert self.solution.pipe_classes("Pipe2_ret")[0].name == "DN150"  # initially DN->None
        assert self.solution.pipe_classes("Pipe2_ret")[-1].name == "DN250"  # initially DN450
        # Check the minimum velocity setting==default value. Keep the default value hard-coded to
        # prevent future coding bugs
        np.testing.assert_equal(1.0e-4, self.solution.heat_network_settings["minimum_velocity"])

        # Check whether cyclic ates constraint is working and split between the charging and
        # discharging heat_flow variables
        for a in self.solution.energy_system_components.get("ates", []):
            stored_heat = self.results[f"{a}.Stored_heat"]
            np.testing.assert_allclose(stored_heat[0], stored_heat[-1], atol=1.0)
            heat_ates = self.results[f"{a}.Heat_ates"]
            heat_flow_charging = self.results[f"{a}.Heat_flow_charging"]
            heat_flow_discharging = self.results[f"{a}.Heat_flow_discharging"]
            np.testing.assert_allclose(
                heat_ates, heat_flow_charging - heat_flow_discharging, atol=1.0
            )
            np.testing.assert_array_less(heat_flow_charging[heat_flow_discharging > 1e3], 1e3)

        # Check variable operational cost of ates
        timesteps_hr = np.diff(self.solution.times()) / 3600.0
        variable_operational_cost = 0.0
        for a in self.solution.energy_system_components.get("ates", []):
            esdl_asset = self.solution.esdl_assets[self.solution.esdl_asset_name_to_id_map[f"{a}"]]
            costs_esdl_asset = esdl_asset.attributes["costInformation"]
            var_op_costs = costs_esdl_asset.variableOperationalCosts.value / 1.0e6  # EUR/Wh_th
            self.assertNotEqual(0.0, var_op_costs)
            for ii in range(1, len(self.solution.times())):
                variable_operational_cost += (
                    var_op_costs
                    * (
                        self.results[f"{a}.Heat_flow_charging"][ii]
                        + self.results[f"{a}.Heat_flow_discharging"][ii]
                    )
                    * timesteps_hr[ii - 1]
                )
        np.testing.assert_allclose(
            variable_operational_cost, self.results[f"{a}__variable_operational_cost"]
        )

        # Check whether buffer tank is only active in peak day
        peak_day_indx = self.solution.parameters(0)["peak_day_index"]
        for b in self.solution.energy_system_components.get("heat_buffer", []):
            heat_buffer = self.results[f"{b}.Heat_buffer"]
            for i in range(len(self.solution.times())):
                if i < peak_day_indx or i > (peak_day_indx + 23):
                    np.testing.assert_allclose(heat_buffer[i], 0.0, atol=1.0e-6)
            heat_flow_charging = self.results[f"{b}.Heat_flow_charging"]
            heat_flow_discharging = self.results[f"{b}.Heat_flow_discharging"]
            np.testing.assert_allclose(
                heat_buffer, heat_flow_charging - heat_flow_discharging, atol=1.0
            )

        obj = self.calculate_objective_value_end_scenario_sizing_all_optional(self.solution)
        excluded_costs_in_obj = self.calculate_heat_demand_costs_end_scenario_sizing()  # Fixed
        # costs excluded in the optim objective function

        # Since all assets are optional, the objective value should be close to the calculated tco
        # excluding heating demands.
        np.testing.assert_allclose(obj / 1.0e6, self.solution.objective_value)
        np.testing.assert_array_less(
            self.solution.objective_value, (obj + excluded_costs_in_obj) / 1.0e6
        )

    def test_end_scenario_sizing_staged(self):
        """
        Check if the EndScenarioSizingStagedHIGHS workflow is behaving as expected. This is an
        optimization done over a full year with timesteps of 5 days and hour timesteps for the peak
        day.

        Checks:
        - Cyclic behaviour for ATES
        - ATES is placed and that the size matches a single doublet, which is larger than the max
        heat_flow
        - That buffer tank is placed and only used on peak day
        - demand matching
        - Check if TCO goal included the desired cost components.

        - Compare objective value of staged approach wit non-staged approach

        Compare solution time of 3 scenarios on how to run the optimisation:
        - Staged approach should be solved faster than the unstaged approaches
        - Unstaged approaches, using the general function run_optimization_problem and the
        function run_end_scenario_sizing with staged_pipe_optimization to False should have
        comparable computation times.


        Missing:
        - Link ATES t0 utilization to state of charge at end of year for optimizations over one
        year.
        """
        import models.test_case_small_network_ates_buffer_optional_assets.src.run_ates as run_ates

        base_folder = Path(run_ates.__file__).resolve().parent.parent

        # This is an optimization done over a full year with timesteps of 5 days and hour timesteps
        # for the peak day

        solution_unstaged = self.solution

        solution_unstaged_2 = run_end_scenario_sizing(
            EndScenarioSizing,
            staged_pipe_optimization=False,
            base_folder=base_folder,
            esdl_file_name="test_case_small_network_with_ates_with_buffer_all_optional.esdl",
            esdl_parser=ESDLFileParser,
            profile_reader=ProfileReaderFromFile,
            input_timeseries_file="Warmte_test.csv",
        )

        solution_staged = run_end_scenario_sizing(
            EndScenarioSizingStaged,
            base_folder=base_folder,
            esdl_file_name="test_case_small_network_with_ates_with_buffer_all_optional.esdl",
            esdl_parser=ESDLFileParser,
            profile_reader=ProfileReaderFromFile,
            input_timeseries_file="Warmte_test.csv",
        )

        results = solution_staged.extract_results()

        # Check whehter the heat demand is matched
        demand_matching_test(solution_staged, results)

        # Check whether cyclic ates constraint is working
        for a in solution_staged.energy_system_components.get("ates", []):
            stored_heat = results[f"{a}.Stored_heat"]
            np.testing.assert_allclose(stored_heat[0], stored_heat[-1], atol=1.0)
        # Check that the ATES is placed and that the size should match the single_doublet_power
        np.testing.assert_allclose(
            results[f"{a}__max_size"],
            self.solution.parameters(0)[f"{a}.single_doublet_power"],
        )
        np.testing.assert_array_less(max(results[f"{a}.Heat_flow"]), results[f"{a}__max_size"])
        np.testing.assert_allclose(results[f"{a}_aggregation_count"], 1)

        # Check whether buffer tank placed and that it is only active in peak day
        peak_day_indx = int(solution_staged.parameters(0)["peak_day_index"])
        for b in solution_staged.energy_system_components.get("heat_buffer", []):
            np.testing.assert_allclose(results[f"{b}_aggregation_count"], 1)  # being placed
            np.testing.assert_array_less(
                1.0e3, max(results[f"{b}.Heat_flow"][peak_day_indx : peak_day_indx + 24])
            )  # at least 1 time step with such a heat_flow is expected in this network
            heat_buffer = results[f"{b}.Heat_buffer"]
            for i in range(len(solution_staged.times())):
                if i < peak_day_indx or i > (peak_day_indx + 23):
                    np.testing.assert_allclose(heat_buffer[i], 0.0, atol=1.0e-6)

        obj = self.calculate_objective_value_end_scenario_sizing_all_optional(solution_staged)
        np.testing.assert_allclose(obj / 1.0e6, solution_staged.objective_value)

        # comparing results of staged and unstaged problem definition. For larger systems there
        # might be a difference in the value but that would either be a difference within the
        # MIPgap (thus checking best bound objective is still smaller or equal than objective of
        # the other problem) or because of some tighter constraints in the staged problem e.g.
        # staged problem slightly higher objective value
        if (
            solution_staged.solver_stats["mip_gap"] == 0.0
            and solution_unstaged.solver_stats["mip_gap"] == 0.0
        ):
            np.testing.assert_allclose(
                solution_staged.objective_value, solution_unstaged.objective_value
            )
        else:
            np.testing.assert_array_less(
                solution_unstaged._priorities_output[1][4]["mip_dual_bound"] - 1e-6,
                solution_staged.objective_value,
            )
            np.testing.assert_array_less(
                solution_staged._priorities_output[3][4]["mip_dual_bound"] - 1e-6,
                solution_unstaged.objective_value,
            )
        # checking time spend on optimisation approaches, the difference between the two unstaged
        # approaches should be smaller than the difference with the staged and unstaged approach.
        # The staged approach should be quickest in solving. The two unstaged approaches should
        # have comparable computation time.
        solution_time_unstaged = sum([i[1] for i in solution_unstaged._priorities_output])
        solution_time_unstaged_2 = sum([i[1] for i in solution_unstaged_2._priorities_output])
        solution_time_staged = sum([i[1] for i in solution_staged._priorities_output])
        np.testing.assert_array_less(
            abs(solution_time_unstaged_2 - solution_time_unstaged),
            abs(solution_time_staged - solution_time_unstaged),
        )
        np.testing.assert_array_less(solution_time_staged, solution_time_unstaged)

    def test_end_scenario_sizing_discounted(self):
        """
        Check if the TestEndScenario sizing workflow is behaving as expected. This is an
        optimization done over a full year with timesteps of 5 days and hour timesteps for the peak
        day.

        Checks:
        - demand matching

        Missing:
        - Check if TCO goal included the desired cost components.

        """
        import models.test_case_small_network_ates_buffer_optional_assets.src.run_ates as run_ates

        base_folder = Path(run_ates.__file__).resolve().parent.parent

        class TestEndScenarioSizingDiscountedHIGHS(EndScenarioSizingDiscounted):
            def solver_options(self):
                options = super().solver_options()
                options["solver"] = "highs"
                highs_options = options["highs"] = {}
                highs_options["mip_rel_gap"] = 0.05
                return options

        # This is an optimization done over a full year with timesteps of 5 days and hour timesteps
        # for the peak day
        solution = run_optimization_problem(
            TestEndScenarioSizingDiscountedHIGHS,
            base_folder=base_folder,
            esdl_file_name="test_case_small_network_all_optional.esdl",
            esdl_parser=ESDLFileParser,
            profile_reader=ProfileReaderFromFile,
            input_timeseries_file="Warmte_test.csv",
            error_type_check=NetworkErrors.NO_POTENTIAL_ERRORS_CHECK,  # Pass the error type here
        )

        results = solution.extract_results()
        # In the future we want to check the following
        # Is the timeline correctly converted, correct peak day, correct amount of timesteps, etc.
        # Check whether expected assets are disabled
        # Check the optimal size of assets
        # Check the cost breakdown, check whether all the enabled assets are in the cost breakdown
        # Check that computation time is within expected bounds

        # Check whether the heat demand is matched
        demand_matching_test(solution, results)

    def test_end_scenario_sizing_head_loss(self):
        """
        Test is EndScenarioSizingHeadLoss class is behaving as expected. E.g. should behave
        similarly to EndScenarioSizing class but now the linearised inequality Darcy Weisbach
        equations should be included for the head loss.

        Checks:
        - objective (TCO) same order of magnitude
        - head loss is calculated, currently only checked if it is equal or larger than the
        DW head_loss
        """

        import models.test_case_small_network_ates_buffer_optional_assets.src.run_ates as run_ates

        base_folder = Path(run_ates.__file__).resolve().parent.parent

        class EndScenarioSizingHeadLossStagedNLines(EndScenarioSizingHeadLossStaged):
            def __init__(self, *args, **kwargs):
                super().__init__(*args, **kwargs)

                self.heat_network_settings["n_linearization_lines"] = 3

        solution = run_end_scenario_sizing(
            EndScenarioSizingHeadLossStagedNLines,
            base_folder=base_folder,
            esdl_file_name="test_case_small_network_all_optional.esdl",
            esdl_parser=ESDLFileParser,
            profile_reader=ProfileReaderFromFile,
            input_timeseries_file="Warmte_test.csv",
            error_type_check=NetworkErrors.NO_POTENTIAL_ERRORS_CHECK,  # Pass the error type here
        )

        results = solution.extract_results()

        demand_matching_test(solution, results)

        tol = 1.0e-9
        pipes = solution.energy_system_components.get("heat_pipe")
        for pipe in pipes:
            pipe_diameter = solution.parameters(0)[f"{pipe}.diameter"]
            pipe_wall_roughness = solution.energy_system_options()["wall_roughness"]
            temperature = solution.parameters(0)[f"{pipe}.temperature"]
            pipe_length = solution.parameters(0)[f"{pipe}.length"]
            if pipe_diameter > 0.0:
                velocities = results[f"{pipe}.Q"] / solution.parameters(0)[f"{pipe}.area"]
            else:
                velocities = results[f"{pipe}.Q"]  # should be zero
            for ii in range(len(results[f"{pipe}.dH"])):
                if velocities[ii] > 0 and pipe_diameter > 0:
                    np.testing.assert_array_less(
                        darcy_weisbach.head_loss(
                            velocities[ii],
                            pipe_diameter,
                            pipe_length,
                            pipe_wall_roughness,
                            temperature,
                        ),
                        abs(results[f"{pipe}.dH"][ii]) + tol,
                    )

    def test_end_scenario_sizing_pipe_catalog(self):
        """
        Checks that the problem uses different pipe costs when a pipe catalog is provided along
        with the ESDL.
        Test 1: Check that the solution pipe classes used were of the pipe measures
        Test 2: Check that the pipe classes generated by instantiating the same problem with
        # different esdl that does not have measures will actually use the original EDR
        # pipe classes. Then we are sure that (rest of the) EDR pipe classes have not been used
        Test 3: If the costs that are used in the problem match the costs from the ESDL Measures

        """

        import models.test_case_small_network_ates_buffer_optional_assets.src.run_ates as run_ates

        base_folder = Path(run_ates.__file__).resolve().parent.parent

        # This is an optimization done over a few days
        solution = run_end_scenario_sizing(
            EndScenarioSizing,
            base_folder=base_folder,
            esdl_file_name="test_case_small_network_all_optional_pipe_catalog.esdl",
            esdl_parser=ESDLFileParser,
            profile_reader=ProfileReaderFromFile,
            input_timeseries_file="Warmte_test.csv",
            error_type_check=NetworkErrors.NO_POTENTIAL_ERRORS_CHECK,
        )

        results = solution.extract_results()
        parameters = solution.parameters(0)

        # Test 1: Check if the solution pipe classes used were of the pipe measures
        filter_type = "Pipe"
        pipe_measures = solution.filter_asset_measures(solution._esdl_measures, filter_type)
        pipe_diameter_cost_map = {
            str(pipe.diameter): pipe.costInformation.investmentCosts.value
            for pipe in pipe_measures.values()
        }
        solution_pipe_classes = solution.get_unique_pipe_classes()
        solution_pipe_class_cost_map = {
            sol_pipe.name: sol_pipe.investment_costs
            for sol_pipe in solution_pipe_classes
            if not sol_pipe.name == "None"
        }
        np.testing.assert_equal(
            solution_pipe_class_cost_map.items() <= pipe_diameter_cost_map.items(), True
        )

        # Test 2: Check that the pipe classes generated by instantiating the same problem with
        # different esdl that does not have measures will actually use the original EDR
        # pipe classes. Then we are sure that (rest of the) EDR pipe classes have not been used
        model_folder = base_folder / "model"
        input_folder = base_folder / "input"
        original_problem = EndScenarioSizingStaged(
            esdl_parser=ESDLFileParser,
            base_folder=base_folder,
            model_folder=model_folder,
            input_folder=input_folder,
            esdl_file_name="test_case_small_network_with_ates_with_buffer_all_optional.esdl",
            profile_reader=ProfileReaderFromFile,
            input_timeseries_file="Warmte_test.csv",
            error_type_check=NetworkErrors.NO_POTENTIAL_ERRORS_CHECK,
        )
        original_problem.pre()
        original_problem_pipe_classes = original_problem.get_unique_pipe_classes()
        original_problem_pipe_class_cost_map = {
            org_pipe.name: org_pipe.investment_costs
            for org_pipe in original_problem_pipe_classes
            if not org_pipe.name == "None"
        }
        np.testing.assert_equal(
            solution_pipe_class_cost_map.items() != original_problem_pipe_class_cost_map.items(),
            True,
        )

        # Test 3: If the costs that are used in the problem match the costs from the ESDL Measures
        # Optimized pipe class map of diameter
        optimized_pipe_classes_dia_map = {
            solution.get_optimized_pipe_class(pipe)
            .inner_diameter: solution.get_optimized_pipe_class(pipe)
            .name
            for pipe in solution.hot_pipes
        }
        # Assert that the same specific investment costs have been use between the measure and the
        # solution parameters
        for heat_pipe in solution.energy_system_components.get("heat_pipe"):
            # Check only if the pipe exits. Cost of 0 means the asset has been removed as a result
            # of the sizing optimization
            if results[f"{solution._asset_investment_cost_map[heat_pipe]}"] > 1.0:
                investment_cost_specific = (
                    results[f"{solution._asset_investment_cost_map[heat_pipe]}"][0]
                    / parameters[f"{heat_pipe}.length"]
                )  # [Eur/m]
                optimized_diameter = parameters[f"{heat_pipe}.diameter"]
                cost_map_from_measures = pipe_diameter_cost_map[
                    optimized_pipe_classes_dia_map[optimized_diameter]
                ]
                np.testing.assert_allclose(cost_map_from_measures, investment_cost_specific)

    def test_end_scenario_sizing_pipe_catalog_with_templates(self):
        """
        This test is a temporary. This checks that current implementation that reads and
        processes pipe measures can also reads and processes pipe templates.
        """

        import models.test_case_small_network_ates_buffer_optional_assets.src.run_ates as run_ates

        base_folder = Path(run_ates.__file__).resolve().parent.parent

        # This is an optimization done over a few days
        solution = run_end_scenario_sizing(
            EndScenarioSizing,
            base_folder=base_folder,
            esdl_file_name="test_case_small_network_all_optional_pipe_catalog_with_templates.esdl",
            esdl_parser=ESDLFileParser,
            profile_reader=ProfileReaderFromFile,
            input_timeseries_file="Warmte_test.csv",
            error_type_check=NetworkErrors.NO_POTENTIAL_ERRORS_CHECK,
        )

        results = solution.extract_results()
        parameters = solution.parameters(0)

        # Test 1: Check if the solution pipe classes used were of the pipe measures
        pipe_measures = solution.filter_asset_measures(solution._esdl_measures, "Pipe")
        pipe_diameter_cost_map = {
            str(pipe.diameter): pipe.costInformation.investmentCosts.value
            for pipe in pipe_measures.values()
        }
        solution_pipe_classes = solution.get_unique_pipe_classes()
        solution_pipe_class_cost_map = {
            sol_pipe.name: sol_pipe.investment_costs
            for sol_pipe in solution_pipe_classes
            if not sol_pipe.name == "None"
        }
        np.testing.assert_equal(
            solution_pipe_class_cost_map.items() <= pipe_diameter_cost_map.items(), True
        )

        # Test 2: Check that the pipe classes generated by instantiating the same problem with
        # different esdl that does not have measures will actually use the original EDR
        # pipe classes. Then we are sure that (rest of the) EDR pipe classes have not been used
        model_folder = base_folder / "model"
        input_folder = base_folder / "input"
        original_problem = EndScenarioSizingStaged(
            esdl_parser=ESDLFileParser,
            base_folder=base_folder,
            model_folder=model_folder,
            input_folder=input_folder,
            esdl_file_name="test_case_small_network_with_ates_with_buffer_all_optional.esdl",
            profile_reader=ProfileReaderFromFile,
            input_timeseries_file="Warmte_test.csv",
            error_type_check=NetworkErrors.NO_POTENTIAL_ERRORS_CHECK,
        )
        original_problem.pre()
        original_problem_pipe_classes = original_problem.get_unique_pipe_classes()
        original_problem_pipe_class_cost_map = {
            org_pipe.name: org_pipe.investment_costs
            for org_pipe in original_problem_pipe_classes
            if not org_pipe.name == "None"
        }
        np.testing.assert_equal(
            solution_pipe_class_cost_map.items() != original_problem_pipe_class_cost_map.items(),
            True,
        )

        # Test 3: If the costs that are used in the problem match the costs from the ESDL Measures
        # Optimized pipe class map of diameter
        optimized_pipe_classes_dia_map = {
            solution.get_optimized_pipe_class(pipe)
            .inner_diameter: solution.get_optimized_pipe_class(pipe)
            .name
            for pipe in solution.hot_pipes
        }
        # Assert that the same specific investment costs have been use between the measure and the
        # solution parameters
        for heat_pipe in solution.energy_system_components.get("heat_pipe"):
            # Check only if the pipe exits. Cost of 0 means the asset has been removed as a result
            # of the sizing optimization
            if results[f"{solution._asset_investment_cost_map[heat_pipe]}"] > 1.0:
                investment_cost_specific = (
                    results[f"{solution._asset_investment_cost_map[heat_pipe]}"][0]
                    / parameters[f"{heat_pipe}.length"]
                )  # [Eur/m]
                optimized_diameter = parameters[f"{heat_pipe}.diameter"]
                cost_map_from_measures = pipe_diameter_cost_map[
                    optimized_pipe_classes_dia_map[optimized_diameter]
                ]
                np.testing.assert_allclose(cost_map_from_measures, investment_cost_specific)

    def calculate_heat_demand_costs_end_scenario_sizing(self):
        excluded_costs_in_obj = 0.0  # Fixed costs excluded in the optim objective function
        years = self.solution.parameters(0)["number_of_years"]
        for asset in [*self.solution.energy_system_components.get("heat_demand", [])]:
            technical_lifetime = self.solution.parameters(0)[f"{asset}.technical_life"]
            factor = years / technical_lifetime
            if factor < 1.0:
                factor = 1.0
            if asset in [*self.solution.energy_system_components.get("heat_demand", [])]:
                excluded_costs_in_obj += (
                    self.results[f"{self.solution._asset_installation_cost_map[asset]}"] * factor
                )
        return excluded_costs_in_obj

    def calculate_objective_value_end_scenario_sizing_all_optional(self, solution):
        # If heating demand asset's state is enabled, then exclude the costs since it is not
        # part of the TCO calculation. This is because we do not size heating demand assets in
        # the optimization
        results = solution.extract_results()
        obj = 0.0
        years = solution.parameters(0)["number_of_years"]
        for asset in [
            *solution.energy_system_components.get("heat_source", []),
            *solution.energy_system_components.get("ates", []),
            *solution.energy_system_components.get("heat_buffer", []),
            *solution.energy_system_components.get("heat_demand", []),
            *solution.energy_system_components.get("heat_exchanger", []),
            *solution.energy_system_components.get("heat_pump", []),
            *solution.energy_system_components.get("heat_pipe", []),
        ]:
            asset_id = solution.esdl_asset_name_to_id_map[asset]
            asset_state = solution.esdl_assets[asset_id].attributes["state"]
            asset_type = solution.esdl_assets[asset_id].asset_type
            if not (
                (asset_type == "HeatingDemand") and (asset_state == esdl.AssetStateEnum.ENABLED)
            ):
                technical_lifetime = solution.parameters(0)[f"{asset}.technical_life"]
                factor = years / technical_lifetime
                if factor < 1.0:
                    factor = 1.0
                obj += results[f"{solution._asset_fixed_operational_cost_map[asset]}"] * years
                obj += results[f"{solution._asset_variable_operational_cost_map[asset]}"] * years
                obj += results[f"{solution._asset_investment_cost_map[asset]}"] * factor
                obj += results[f"{solution._asset_installation_cost_map[asset]}"] * factor

        return obj


if __name__ == "__main__":
    import time

    start_time = time.time()
    a = TestEndScenarioSizing()
    a.setUpClass()
    a.test_end_scenario_sizing()
    a.test_end_scenario_sizing_staged()
    a.test_end_scenario_sizing_discounted()
    a.test_end_scenario_sizing_head_loss()
    a.test_end_scenario_sizing_pipe_catalog()
    a.test_end_scenario_sizing_pipe_catalog_with_templates()
    print("Execution time: " + time.strftime("%M:%S", time.gmtime(time.time() - start_time)))



from enum import Enum
from typing import Dict, List, NoReturn

from mesido.exceptions import MesidoAssetIssueError

AssetId = str
ErrorMessage = str


# Asset error type that can occur
class MesidoAssetIssueType(Enum):
    """
    Enumeration of possible asset issue types in the MESIDO system.

    This enum defines the different types of issues that can occur with assets in the system.
    Each issue type has a specific validation purpose:

    - HEAT_PRODUCER_POWER, Validate if power producers have sufficient installed capacity
    - HEAT_DEMAND_POWER, COLD_DEMAND_POWER: Validate that installed capacity meets demand
    - HEAT_DEMAND_TYPE: Validate correct asset type for heat demand
    - ASSET_PROFILE_CAPABILITY: Validate profile assignment capabilities
    - ASSET_COST_ATTRIBUTE_INCORRECT: Validate that existing cost attributes have
        correct values/units
    - ASSET_COST_ATTRIBUTE_MISSING: Validate that required cost attributes are present
    - HEAT_EXCHANGER_TEMPERATURES: Validate temperature settings for heat exchangers

    For cost attributes:
    Cost attributes can be required or optional for each asset.
    - Required inputs: Must exist in the ESDL file (ASSET_COST_ATTRIBUTE_MISSING) and
      have correct values/units (ASSET_COST_ATTRIBUTE_INCORRECT)
    - Optional inputs: Not required to exist, but if present must have correct
      values/units (ASSET_COST_ATTRIBUTE_INCORRECT)
    """

    HEAT_PRODUCER_POWER = "heat_producer.power"
    HEAT_DEMAND_POWER = "heat_demand.power"
    COLD_DEMAND_POWER = "cold_demand.power"
    HEAT_DEMAND_TYPE = "heat_demand.type"
    ASSET_PROFILE_CAPABILITY = "asset_profile.capability"
    ASSET_PROFILE_AVAILABILITY = "asset_profile.availability"
    ASSET_COST_ATTRIBUTE_INCORRECT = "asset_cost_attribute.incorrect"
    ASSET_COST_ATTRIBUTE_MISSING = "asset_cost_attribute.missing"
    HEAT_EXCHANGER_TEMPERATURES = "heat_exchanger.temperature"
    HEAT_EXCHANGER_POWER = "heat_exchanger.capacity"
    HEAT_DEMAND_STATE = "heat_demand.state"
    ASSET_PROFILE_MULTIPLIER = "asset_profile.multiplier"
    COLD_ASSET_TYPE_NOT_SUPPORTED = "cold_asset.type"
    ELECT_ASSET_TYPE_NOT_SUPPORTED = "elect_asset.type"


class PotentialErrors:
    """Singleton, do not instantiate. Use POTENTIAL_ERRORS."""

    _gathered_potential_issues: Dict[MesidoAssetIssueType, Dict[AssetId, List[ErrorMessage]]]

    def __init__(self):
        self._gathered_potential_issues = {}

    def add_potential_issue(
        self, issue_type: MesidoAssetIssueType, asset_id: AssetId, error_message: ErrorMessage
    ) -> None:
        """
        Add potential issues to _gathered_potential_issues.

        """
        self._gathered_potential_issues.setdefault(issue_type, {})
        self._gathered_potential_issues[issue_type][asset_id] = error_message

    def have_issues_for(self, issue_type: MesidoAssetIssueType) -> bool:
        """
        Check if the potential issue exists.

        """
        result = False
        if issue_type in self._gathered_potential_issues:
            result = True

        return result

    def convert_to_exception(
        self, issue_type: MesidoAssetIssueType, general_issue: str
    ) -> NoReturn:
        """
        Raise a MESIDO exception if the issue exists.

        """
        if issue_type not in self._gathered_potential_issues:
            raise RuntimeError("Something very wrong. Issue type not in potential errors")

        raise MesidoAssetIssueError(
            general_issue=general_issue,
            error_type=issue_type,
            message_per_asset_id=self._gathered_potential_issues[issue_type],
        )


# When adding POTENTIAL_ERRORS to a workflow a reset thereof is required due to it being a
# persistent object
POTENTIAL_ERRORS = PotentialErrors()


def get_potential_errors() -> PotentialErrors:
    return POTENTIAL_ERRORS


def reset_potential_errors() -> None:
    POTENTIAL_ERRORS._gathered_potential_issues = {}



import datetime
import unittest
import unittest.mock
from pathlib import Path
from typing import Optional

import esdl

from mesido.esdl.esdl_parser import ESDLFileParser
from mesido.esdl.profile_parser import InfluxDBProfileReader
from mesido.exceptions import MesidoAssetIssueError
from mesido.potential_errors import MesidoAssetIssueType, PotentialErrors
from mesido.workflows import EndScenarioSizingStaged
from mesido.workflows.utils.error_types import mesido_issue_type_gen_message

import numpy as np

import pandas as pd

from utils_test_scaling import create_log_list_scaling


class MockInfluxDBProfileReader(InfluxDBProfileReader):
    def __init__(
        self,
        energy_system: esdl.EnergySystem,
        file_path: Optional[Path],
    ):
        super().__init__(
            energy_system=energy_system,
            file_path=file_path,
        )
        self._loaded_profiles = pd.read_csv(
            file_path,
            index_col="DateTime",
            date_parser=lambda x: pd.to_datetime(x).tz_convert(datetime.timezone.utc),
        )

    def _load_profile_timeseries_from_database(self, profile: esdl.InfluxDBProfile) -> pd.Series:
        return self._loaded_profiles[profile.id]


class TestPotentialErrors(unittest.TestCase):
    def test_asset_potential_errors(self):
        """
        This test checks that the error checks in the code for sufficient installed cool/heatig
        capacity of a cold/heat demand is sufficient (grow_workflow)

        Checks:
        1. Correct error is raised
        2. That the error is due to:
            - insufficient heat specified capacities for 3 heating demands
            - incorrect heating demand type being used for 1 heating demand
            - profile cannot be assigned to a specific asset
            - state set to optional for 2 assets, since that is not allowed
        """
        import models.unit_cases.case_1a.src.run_1a as run_1a

        base_folder = Path(run_1a.__file__).resolve().parent.parent
        model_folder = base_folder / "model"
        input_folder = base_folder / "input"

        logger, logs_list = create_log_list_scaling("mesido")

        with (
            self.assertRaises(MesidoAssetIssueError) as cm,
            unittest.mock.patch("mesido.potential_errors.POTENTIAL_ERRORS", PotentialErrors()),
        ):
            problem = EndScenarioSizingStaged(
                esdl_parser=ESDLFileParser,
                base_folder=base_folder,
                model_folder=model_folder,
                input_folder=input_folder,
                esdl_file_name="1a_with_influx_profiles_error_check_1.esdl",
                profile_reader=MockInfluxDBProfileReader,
                input_timeseries_file="influx_mock.csv",
            )
            problem.pre()

        # Check that the heat demand had an error
        np.testing.assert_equal(cm.exception.error_type, MesidoAssetIssueType.HEAT_DEMAND_POWER)
        np.testing.assert_equal(
            cm.exception.general_issue,
            mesido_issue_type_gen_message(MesidoAssetIssueType.HEAT_DEMAND_POWER),
        )
        np.testing.assert_equal(
            cm.exception.message_per_asset_id["2ab92324-f86e-4976-9a6e-f7454b77ba3c"],
            "Asset named HeatingDemand_2ab9: The installed capacity of 6.0MW should be larger than"
            " the maximum of the heat demand profile 5175.717MW",
        )
        np.testing.assert_equal(
            cm.exception.message_per_asset_id["506c41ac-d415-4482-bf10-bf12f17aeac6"],
            "Asset named HeatingDemand_506c: The installed capacity of 2.0MW should be larger than"
            " the maximum of the heat demand profile 1957.931MW",
        )
        np.testing.assert_equal(
            cm.exception.message_per_asset_id["6662aebb-f85e-4df3-9f7e-c58993586fba"],
            "Asset named HeatingDemand_6662: The installed capacity of 2.0MW should be larger than"
            " the maximum of the heat demand profile 1957.931MW",
        )
        np.testing.assert_equal(len(cm.exception.message_per_asset_id), 3.0)

        # Check heating demand type error
        with (
            self.assertRaises(MesidoAssetIssueError) as cm,
            unittest.mock.patch("mesido.potential_errors.POTENTIAL_ERRORS", PotentialErrors()),
        ):
            problem = EndScenarioSizingStaged(
                esdl_parser=ESDLFileParser,
                base_folder=base_folder,
                model_folder=model_folder,
                input_folder=input_folder,
                esdl_file_name="1a_with_influx_profiles_error_check_2.esdl",
                profile_reader=MockInfluxDBProfileReader,
                input_timeseries_file="influx_mock.csv",
            )
            problem.pre()
        # Check that the heat demand had an error
        np.testing.assert_equal(cm.exception.error_type, MesidoAssetIssueType.HEAT_DEMAND_TYPE)
        np.testing.assert_equal(
            cm.exception.general_issue,
            mesido_issue_type_gen_message(MesidoAssetIssueType.HEAT_DEMAND_TYPE),
        )
        np.testing.assert_equal(
            cm.exception.message_per_asset_id["2ab92324-f86e-4976-9a6e-f7454b77ba3c"],
            "Asset named HeatingDemand_2ab9: This asset is currently a GenericConsumer please"
            " change it to a HeatingDemand",
        )
        np.testing.assert_equal(len(cm.exception.message_per_asset_id), 1.0)

        # Check asset profile capability
        with (
            self.assertRaises(MesidoAssetIssueError) as cm,
            unittest.mock.patch("mesido.potential_errors.POTENTIAL_ERRORS", PotentialErrors()),
        ):
            problem = EndScenarioSizingStaged(
                esdl_parser=ESDLFileParser,
                base_folder=base_folder,
                model_folder=model_folder,
                input_folder=input_folder,
                esdl_file_name="1a_with_influx_profiles_error_check_3.esdl",
                profile_reader=MockInfluxDBProfileReader,
                input_timeseries_file="influx_mock.csv",
            )
            problem.pre()
        # Check that the joint has an error
        np.testing.assert_equal(
            cm.exception.error_type,
            MesidoAssetIssueType.ASSET_PROFILE_CAPABILITY,
        )
        np.testing.assert_equal(
            cm.exception.general_issue,
            mesido_issue_type_gen_message(MesidoAssetIssueType.ASSET_PROFILE_CAPABILITY),
        )
        np.testing.assert_equal(
            cm.exception.message_per_asset_id["95802cf8-61d6-4773-bb99-e275c3bf26cc"],
            "Asset named Joint_9580: The assigment of profile field demand3_MW is not possible for"
            " this asset type <class 'esdl.esdl.Joint'>",
        )
        np.testing.assert_equal(len(cm.exception.message_per_asset_id), 1.0)

        # Check that the heating demand is set to optional
        with (
            self.assertRaises(MesidoAssetIssueError) as cm,
            unittest.mock.patch("mesido.potential_errors.POTENTIAL_ERRORS", PotentialErrors()),
        ):
            problem = EndScenarioSizingStaged(
                esdl_parser=ESDLFileParser,
                base_folder=base_folder,
                model_folder=model_folder,
                input_folder=input_folder,
                esdl_file_name="1a_with_influx_profiles_error_check_4.esdl",
                profile_reader=MockInfluxDBProfileReader,
                input_timeseries_file="influx_mock.csv",
            )
            problem.pre()
        # Check that the heat demand had an error
        np.testing.assert_equal(cm.exception.error_type, MesidoAssetIssueType.HEAT_DEMAND_STATE)
        np.testing.assert_equal(
            cm.exception.general_issue,
            mesido_issue_type_gen_message(MesidoAssetIssueType.HEAT_DEMAND_STATE),
        )
        np.testing.assert_equal(
            cm.exception.message_per_asset_id["2ab92324-f86e-4976-9a6e-f7454b77ba3c"],
            "Asset named HeatingDemand_2ab9 : The asset should be enabled since there is "
            "no sizing optimization on HeatingDemands",
        )
        np.testing.assert_equal(
            cm.exception.message_per_asset_id["6662aebb-f85e-4df3-9f7e-c58993586fba"],
            "Asset named HeatingDemand_6662 : The asset should be enabled since there is "
            "no sizing optimization on HeatingDemands",
        )
        np.testing.assert_equal(len(cm.exception.message_per_asset_id), 2.0)

        # Check the a new type of potential error which raises when the profile name indicated
        # in esdl is not available in the database
        with (
            self.assertRaises(MesidoAssetIssueError) as cm,
            unittest.mock.patch("mesido.potential_errors.POTENTIAL_ERRORS", PotentialErrors()),
        ):
            problem = EndScenarioSizingStaged(
                esdl_parser=ESDLFileParser,
                base_folder=base_folder,
                model_folder=model_folder,
                input_folder=input_folder,
                esdl_file_name="1a_with_influx_profiles_wrong_name.esdl",
            )
            problem.pre()
        # Check that the asset profile had an error
        np.testing.assert_equal(
            cm.exception.error_type, MesidoAssetIssueType.ASSET_PROFILE_AVAILABILITY
        )
        np.testing.assert_equal(
            cm.exception.message_per_asset_id["2ab92324-f86e-4976-9a6e-f7454b77ba3c"],
            "Asset named HeatingDemand_2ab9: Input profile "
            "demand1_MW_wrong_name in WarmingUp default profiles is not available in the database.",
        )

        # Check that the ResidualHeatSource multiplier's error is picked up
        with (
            self.assertRaises(MesidoAssetIssueError) as cm,
            unittest.mock.patch("mesido.potential_errors.POTENTIAL_ERRORS", PotentialErrors()),
        ):
            problem = EndScenarioSizingStaged(
                esdl_parser=ESDLFileParser,
                base_folder=base_folder,
                model_folder=model_folder,
                input_folder=input_folder,
                esdl_file_name="1a_with_influx_profiles_error_check_5.esdl",
                profile_reader=InfluxDBProfileReader,
            )
            problem.pre()
        np.testing.assert_equal(
            cm.exception.message_per_asset_id["8172d5d3-61a4-4d0b-a26f-5e61c2a22c64"],
            ", GenericProducer_8172 has unit's multiplier specified incorrectly. "
            "Multiplier should be 1.0, when the unit is specified in Coefficient in %",
        )


if __name__ == "__main__":
    a = TestPotentialErrors()
    a.test_asset_potential_errors()



from pathlib import Path
from unittest import TestCase

from mesido.esdl.esdl_parser import ESDLFileParser
from mesido.esdl.profile_parser import ProfileReaderFromFile
from mesido.util import run_esdl_mesido_optimization

import numpy as np

from utils_tests import demand_matching_test, energy_conservation_test, heat_to_discharge_test


class TestLogicalLinks(TestCase):
    def test_logical_links_electricity(self):
        """
        This test checks if the logic implemented for logically linked energy system, Meaning an
        energy system where assets are directly connected to each other without a network
        (transport asset(s)) in between.

        Checks:
        1. Check that energy is conserved

        """
        import models.unit_cases_electricity.bus_networks.src.example as example
        from models.unit_cases_electricity.bus_networks.src.example import ElectricityProblem

        base_folder = Path(example.__file__).resolve().parent.parent

        problem = run_esdl_mesido_optimization(
            ElectricityProblem,
            esdl_file_name="Electric_bus4_logical_links.esdl",
            base_folder=base_folder,
            esdl_parser=ESDLFileParser,
            profile_reader=ProfileReaderFromFile,
            input_timeseries_file="timeseries.csv",
        )
        results = problem.extract_results()

        np.testing.assert_allclose(
            results["ElectricityDemand_281a.Electricity_demand"]
            + results["ElectricityDemand_e527.Electricity_demand"],
            results["ElectricityProducer_17a1.Electricity_source"]
            + results["ElectricityProducer_a215.Electricity_source"],
        )

    def test_logical_links_gas(self):
        """
        This test checks if the logic implemented for logically linked energy system, Meaning an
        energy system where assets are directly connected to each other without a network
        (transport asset(s)) in between.

        Checks:
        1. Check that energy is conserved

        """
        import models.unit_cases_gas.multi_demand_source_node.src.run_test as example
        from models.unit_cases_gas.multi_demand_source_node.src.run_test import GasProblem

        base_folder = Path(example.__file__).resolve().parent.parent

        problem = run_esdl_mesido_optimization(
            GasProblem,
            esdl_file_name="test_logical_links.esdl",
            base_folder=base_folder,
            esdl_parser=ESDLFileParser,
            profile_reader=ProfileReaderFromFile,
            input_timeseries_file="timeseries.csv",
        )
        results = problem.extract_results()

        np.testing.assert_allclose(
            results["GasDemand_47d0.Gas_demand_mass_flow"]
            + results["GasDemand_7978.Gas_demand_mass_flow"],
            results["GasProducer_a977.Gas_source_mass_flow"]
            + results["GasProducer_3573.Gas_source_mass_flow"],
        )

    def test_logical_links_nodes(self):
        """
        This test checks if the logic implemented for a partially logically linked energy system.
        Meaning an energy system where assets are directly connected to each other without a network
        (transport asset(s)) in between. This specific test covers the logical connection between
        two nodes.
        """

        import models.unit_cases_gas.source_pipe_split_sink.src.run_source_sink as example
        from models.unit_cases_gas.source_pipe_split_sink.src.run_source_sink import GasProblem

        base_folder = Path(example.__file__).resolve().parent.parent

        problem = run_esdl_mesido_optimization(
            GasProblem,
            esdl_file_name="split_with_logical_link.esdl",
            base_folder=base_folder,
            esdl_parser=ESDLFileParser,
            profile_reader=ProfileReaderFromFile,
            input_timeseries_file="timeseries.csv",
        )
        results = problem.extract_results()

        demand_matching_test(problem, results)

        # We check that no artificial gas in created
        np.testing.assert_allclose(
            results["GasProducer_0876.GasOut.Q"], results["GasDemand_a2d8.GasIn.Q"]
        )

        # We test conservaiton of flow at the nodes
        for node, connected_pipes in problem.energy_system_topology.gas_nodes.items():
            discharge_sum = 0.0

            for i_conn, (_pipe, orientation) in connected_pipes.items():
                discharge_sum += results[f"{node}.GasConn[{i_conn+1}].Q"] * orientation
                np.testing.assert_allclose(
                    results[f"{node}.GasConn[{i_conn+1}].H"], results[f"{node}.H"], atol=1.0e-6
                )
            np.testing.assert_allclose(discharge_sum, 0.0, atol=1.0e-12)

    def test_logical_links_nodes_heat(self):
        """
        This test checks if the logic implemented for a partially logically linked energy system.
        Meaning an energy system where assets are directly connected to each other without a network
        (transport asset(s)) in between. This specific test covers the logical connection between
        two nodes in a heat network. Please note that due to the closed system constraints we need
        to set the minimum velocity to zero to avoid backflow in the solution.
        """

        import models.source_pipe_split_sink.src.double_pipe_heat as example
        from models.source_pipe_split_sink.src.double_pipe_heat import SourcePipeSink

        base_folder = Path(example.__file__).resolve().parent.parent

        problem = run_esdl_mesido_optimization(
            SourcePipeSink,
            esdl_file_name="sourcesink_with_node2node_logical_link.esdl",
            base_folder=base_folder,
            esdl_parser=ESDLFileParser,
            profile_reader=ProfileReaderFromFile,
            input_timeseries_file="timeseries_import.csv",
        )
        results = problem.extract_results()

        demand_matching_test(problem, results)
        energy_conservation_test(problem, results)
        heat_to_discharge_test(problem, results)

        # We test conservaiton of flow at the nodes
        for node, connected_pipes in problem.energy_system_topology.nodes.items():
            discharge_sum = 0.0

            for i_conn, (_pipe, orientation) in connected_pipes.items():
                discharge_sum += results[f"{node}.HeatConn[{i_conn+1}].Q"] * orientation
                np.testing.assert_allclose(
                    results[f"{node}.HeatConn[{i_conn+1}].H"], results[f"{node}.H"], atol=1.0e-6
                )
            np.testing.assert_allclose(discharge_sum, 0.0, atol=1.0e-12)

    def test_logical_links_network_hybrid(self):
        """
        This test checks if the logic implemented for logically linked energy system, Meaning an
        energy system where assets are directly connected to each other without a network
        (transport asset(s)) in between. This test uses the emerge use-case and the logic is both
        checked for the gas and heat commodity.

        Checks:
        1. Check that energy is conserved

        """
        import models.emerge.src.example as example
        from models.emerge.src.example import EmergeTest

        base_folder = Path(example.__file__).resolve().parent.parent

        problem = run_esdl_mesido_optimization(
            EmergeTest,
            base_folder=base_folder,
            esdl_file_name="emerge_with_logical_links.esdl",
            esdl_parser=ESDLFileParser,
            profile_reader=ProfileReaderFromFile,
            input_timeseries_file="timeseries_short.csv",
        )
        results = problem.extract_results()

        assert all(
            results["ElectricityCable_e388.ElectricityIn.V"]
            == results["Bus_24cf.ElectricityConn[3].V"]
        )

        assert (
            np.sum(
                np.abs(
                    results["WindPark_9074.ElectricityOut.V"]
                    - results["Bus_24cf.ElectricityConn[3].V"]
                )
            )
            != 0.0
        )


if __name__ == "__main__":
    import time

    start_time = time.time()
    a = TestLogicalLinks()
    a.test_logical_links_electricity()
    a.test_logical_links_gas()
    a.test_logical_links_nodes()
    a.test_logical_links_nodes_heat()
    a.test_logical_links_network_hybrid()
    print("Execution time: " + time.strftime("%M:%S", time.gmtime(time.time() - start_time)))



import base64
from pathlib import Path
from typing import Dict, Optional

import esdl.esdl_handler

from mesido.esdl.common import Asset


class _ESDLInputException(Exception):
    pass


class BaseESDLParser:
    def __init__(self):
        self._global_properties: Dict[str, Dict] = {
            "carriers": dict(),
        }
        self._assets: Dict[str, Asset] = dict()
        self._energy_system_handler: esdl.esdl_handler.EnergySystemHandler = (
            esdl.esdl_handler.EnergySystemHandler()
        )
        self._energy_system: Optional[esdl.EnergySystem] = None
        self._esdl_string: Optional[str] = None
        self._esdl_path: Optional[Path] = None
        self._measures: Dict[str, Asset] = dict()

    def _load_esdl_model(self) -> None:
        """
        This function should be implemented by the child. The function doesn't return anything but
        should set the _esdl_model property.
        """
        raise NotImplementedError

    def read_esdl(self) -> None:
        self._load_esdl_model()
        id_to_idnumber_map = {}

        for x in self._energy_system.energySystemInformation.carriers.carrier.items:
            if isinstance(x, esdl.esdl.HeatCommodity):
                if x.id not in id_to_idnumber_map:
                    number_list = [int(s) for s in x.id if s.isdigit()]
                    number = ""
                    for nr in number_list:
                        number = number + str(nr)
                    # note this fix is to create a unique number for the map for when the pipe
                    # duplicator service is used and an additional _ret is added to the id.
                    if "_ret" in x.id:
                        number = number + "000"
                    id_to_idnumber_map[x.id] = int(number)

                temperature = x.supplyTemperature if x.supplyTemperature else x.returnTemperature
                assert temperature > 0.0

                self._global_properties["carriers"][x.id] = dict(
                    name=x.name,
                    id=x.id,
                    id_number_mapping=id_to_idnumber_map[x.id],
                    temperature=temperature,
                    type="milp",
                )
            elif isinstance(x, esdl.esdl.ElectricityCommodity):
                if x.id not in id_to_idnumber_map:
                    number_list = [int(s) for s in x.id if s.isdigit()]
                    number = ""
                    for nr in number_list:
                        number = number + str(nr)
                    id_to_idnumber_map[x.id] = int(number)
                self._global_properties["carriers"][x.id] = dict(
                    name=x.name,
                    voltage=x.voltage,
                    id=x.id,
                    type="electricity",
                    id_number_mapping=id_to_idnumber_map[x.id],
                )
            elif isinstance(x, esdl.esdl.GasCommodity):
                if x.id not in id_to_idnumber_map:
                    number_list = [int(s) for s in x.id if s.isdigit()]
                    number = ""
                    for nr in number_list:
                        number = number + str(nr)
                    id_to_idnumber_map[x.id] = int(number)
                self._global_properties["carriers"][x.id] = dict(
                    name=x.name,
                    pressure=x.pressure,
                    id=x.id,
                    type="gas",
                    id_number_mapping=id_to_idnumber_map[x.id],
                )

        # Component ids are unique, but we require component names to be unique as well.
        component_names = set()

        # Check if the ESDL has asset measures
        try:
            asset_measures = list(self._energy_system.measures.eAllContents())
        except AttributeError:
            asset_measures = None

        # Check if the ESDL has asset templates
        # Currently the use of templates are allowed on a temporary basis, and
        # the templates are stored in the measures variable
        try:
            asset_templates = list(self._energy_system.templates.eAllContents())
        except AttributeError:
            asset_templates = None

        # loop through assets
        for el in self._energy_system.eAllContents():
            # If asset measures exist, collect that in a different dictionary, to be used later in
            # esdl_mixin to update that information
            if (asset_measures is not None and el in asset_measures) or (
                asset_templates is not None and el in asset_templates
            ):
                if isinstance(el, esdl.Measure) or isinstance(el, esdl.AssetTemplate):
                    asset_type = el.__class__.__name__
                    # Note that e.g. el.__dict__['length'] does not work to get the length of a
                    # pipe.
                    # We therefore built this dict ourselves using 'dir' and 'getattr'
                    attributes = {k: getattr(el, k) for k in dir(el)}
                    self._measures[el.id] = Asset(
                        asset_type=asset_type,
                        id=el.id,
                        name=el.name,
                        in_ports=None,
                        out_ports=None,
                        attributes=attributes,
                        global_properties=self._global_properties,
                    )
            else:
                if isinstance(el, esdl.Asset):
                    if hasattr(el, "name") and el.name:
                        el_name = el.name
                    else:
                        el_name = el.id

                    if "." in el_name:
                        # Dots indicate hierarchy, so would be very confusing
                        raise ValueError(f"Dots in component names not supported: '{el_name}'")

                    if el_name in component_names:
                        raise Exception(
                            f"Asset names have to be unique: '{el_name}' already exists"
                        )
                    else:
                        component_names.add(el_name)

                    # For some reason `esdl_element.assetType` is `None`, so use the class name
                    asset_type = el.__class__.__name__

                    # Every asset should at least have a port to be connected to another asset
                    assert len(el.port) >= 1

                    in_ports = None
                    out_ports = None
                    for port in el.port:
                        if isinstance(port, esdl.InPort):
                            if in_ports is None:
                                in_ports = [port]
                            else:
                                in_ports.append(port)
                        elif isinstance(port, esdl.OutPort) and port.name != "EmissionPort":
                            # TODO: need to check if we can have a clearer identifier to the
                            # emissionport such that it is not considered as a connecting port.
                            if out_ports is None:
                                out_ports = [port]
                            else:
                                out_ports.append(port)
                        else:
                            _ESDLInputException(
                                f"The port for {el_name} is neither an IN or OUT port"
                            )

                    # Note that e.g. el.__dict__['length'] does not work to get the length of a
                    # pipe.
                    # We therefore built this dict ourselves using 'dir' and 'getattr'
                    attributes = {k: getattr(el, k) for k in dir(el)}
                    self._assets[el.id] = Asset(
                        asset_type,
                        el.id,
                        el_name,
                        in_ports,
                        out_ports,
                        attributes,
                        self._global_properties,
                    )

    def get_assets(self) -> Dict[str, Asset]:
        return self._assets

    def get_carrier_properties(self) -> Dict:
        return self._global_properties["carriers"]

    def get_esh(self) -> esdl.esdl_handler.EnergySystemHandler:
        return self._energy_system_handler

    def get_measures(self) -> Dict[str, Asset]:
        return self._measures


class ESDLStringParser(BaseESDLParser):
    def __init__(self, **kwargs):
        super().__init__()
        try:
            esdl_string = kwargs.get("esdl_string")
        except KeyError:
            raise _ESDLInputException(
                "Expected an ESDL string when parsing the system from a "
                "string, but none provided"
            )
        if isinstance(esdl_string, bytes):
            self._esdl_string = base64.b64decode(esdl_string).decode("utf-8")
        else:
            self._esdl_string = esdl_string

    def _load_esdl_model(self) -> None:
        self._energy_system = self._energy_system_handler.load_from_string(self._esdl_string)


class ESDLFileParser(BaseESDLParser):
    def __init__(self, **kwargs):
        super().__init__()
        try:
            esdl_path = kwargs.get("esdl_path")
        except KeyError:
            raise _ESDLInputException(
                "Expected an ESDL path when parsing the system from a file but none provided"
            )
        self._esdl_path = esdl_path

    def _load_esdl_model(self) -> None:
        self._energy_system = self._energy_system_handler.load_file(str(self._esdl_path))



from mesido.pycml import Variable
from mesido.pycml.pycml_mixin import add_variables_documentation_automatically

from ._ates_base_asset import _ATESBaseAsset


@add_variables_documentation_automatically
class LowTemperatureATES(_ATESBaseAsset):
    """
    TODO: This model is still under developement.
    A low temperature ates is a underground aquifier in which heat can be stored.

    Like all storage assets we enforce that they must be connected as a demand. The heat to
    discharge constraints are set in the HeatPhysicsMixin, where we use a big_m formulation to
    enforce the correct constraints depending on whether the ates is charging or discharging.

    Please note that:
    The user is responsible to implement the cyclic behaviour in their workflow constraints.
    Meaning that the milp stored at the 1st and last time step should be equal. Furthermore, due
    to the implicit solving note that the energy out of the ATES should be 0 for the 1st time step.

    Variables created:
        {add_variable_names_for_documentation_here}

    Parameters:
        name : The name of the asset. \n
        modifiers : Dictionary with asset information..
    """

    def __init__(self, name, **modifiers):
        super().__init__(name, **modifiers)

        self.component_type = "low_temperature_ates"

        # Stored_heat is the heat that is contained in the ates.
        # Heat_low_temperature_ates is the amount of heat added to or extracted from the buffer
        # per timestep.
        # Thus Heat_buffer = HeatHot = der(Stored_heat).
        # We connect an ATES as an demand, meaning that flow and Heat_low_temperature_ates are
        # positive undercharging and negative under discharge
        self.add_variable(Variable, "Heat_low_temperature_ates", nominal=self.Heat_nominal)  # [W]

        self.add_equation((self.Heat_ates - self.Heat_low_temperature_ates) / self.Heat_nominal)



from mesido.pycml.pycml_mixin import add_variables_documentation_automatically

from numpy import nan

from .gas_base import GasTwoPort
from .._internal import BaseAsset


@add_variables_documentation_automatically
class GasSubstation(GasTwoPort, BaseAsset):
    """
    A gas substation that reduces the pressure level of the flow
    (basically pressure reducinng valve).

    .__disabled

    Variables created:
        {add_variable_names_for_documentation_here}

    Parameters:
        name : The name of the asset. \n
        modifiers : Dictionary with asset information.
    """

    def __init__(self, name, **modifiers):
        super().__init__(name, **modifiers)

        self.component_type = "gas_substation"
        self.min_head = 30.0

        self.Q_nominal_in = nan
        self.Q_nominal_out = nan

        self.density_in = 2.5e3  # H2 density [g/m3] at 30bar
        self.density_out = 2.5e3  # H2 density [g/m3] at 30bar

        self.add_equation(
            (
                (self.GasIn.Q * self.density_in - self.GasOut.Q * self.density_out)
                / (self.Q_nominal_in * self.density_in * self.Q_nominal_out * self.density_out)
                ** 0.5
            )
        )

        self.add_equation(
            ((self.GasIn.Q - self.GasIn.mass_flow / self.density_in) / self.Q_nominal_in)
        )
        self.add_equation(
            ((self.GasOut.Q - self.GasOut.mass_flow / self.density_out) / self.Q_nominal_out)
        )



from mesido.pycml import Variable
from mesido.pycml.component_library.milp.heat.heat_source import HeatSource
from mesido.pycml.pycml_mixin import add_variables_documentation_automatically

from numpy import nan


@add_variables_documentation_automatically
class HeatSourceElec(HeatSource):
    """
    The heat source elec component represents the operational behavior of e-boilers.
    It functions as a heat source that supplies thermal energy to the connected heat network.

    Variables created:
        {add_variable_names_for_documentation_here}

    Parameters:
        name : The name of the asset. \n
        modifiers : Dictionary with asset information.
    """

    def __init__(self, name, **modifiers):
        super().__init__(
            name,
            **modifiers,
        )

        self.efficiency = nan
        self.elec_power_nominal = nan
        self.component_subtype = "heat_source_elec"

        self.add_variable(Variable, "Power_consumed", min=0.0, nominal=self.elec_power_nominal)

        self.add_equation(
            ((self.Power_consumed * self.efficiency - self.Heat_source) / self.Heat_nominal)
        )



import logging
from typing import Dict, List, Set, Tuple

from mesido.base_component_type_mixin import BaseComponentTypeMixin
from mesido.heat_network_common import NodeConnectionDirection
from mesido.network_common import NetworkSettings
from mesido.topology import Topology

import numpy as np

from pymoca.backends.casadi.alias_relation import AliasRelation

logger = logging.getLogger("mesido")


class ModelicaComponentTypeMixin(BaseComponentTypeMixin):
    """
    This class is used to make the milp network component information easily accesible in the
    heat_mixin. This is achieved by creating a heat_network_components dict where the assets are
    sorted by asset_type. Furthermore, the topology object is created in which for specific assets
    the connections with directions are saved for later use in the constraints.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(**kwargs)
        self.__hn_component_types = None
        self.__commodity_types = None

        self.__hot_cold_pipe_relations = dict()
        self.__unrelated_pipes = list()
        self.hot_cold_pipe_relations()

    def pre(self):
        """
        In this function the topology object of the milp network is constructed. Meaning that for
        nodes, busses and storage assets their relevant information on the positive flow direction
        and connections on the ports is gathered and stored in the topology object.
        """
        components = self.energy_system_components
        nodes = components.get("node", [])
        busses = components.get("electricity_node", [])
        gas_nodes = components.get("gas_node", [])
        buffers = components.get("heat_buffer", [])
        atess = [*components.get("ates", []), *components.get("low_temperature_ates", [])]
        demands = [
            *components.get("heat_demand", []),
            *components.get("electricity_demand", []),
            *components.get("gas_demand", []),
        ]
        sources = [
            *components.get("heat_source", []),
            *components.get("electricity_source", []),
            *components.get("gas_source", []),
        ]
        pipes = components.get("heat_pipe", [])

        # An energy system should have at least one asset.
        assert len(components) > 1

        # Figure out which pipes are connected to which nodes, which pipes
        # are connected in series, and which pipes are connected to which buffers.

        pipes_set = set(pipes)
        parameters = [self.parameters(e) for e in range(self.ensemble_size)]
        node_connections = {}
        bus_connections = {}
        gas_node_connections = {}

        heat_network_model_type = NetworkSettings.NETWORK_TYPE_HEAT

        # Note that a pipe series can include both hot and cold pipes for
        # QTH models. It is only about figuring out which pipes are
        # related direction-wise.
        # For Heat models, only hot pipes are allowed to be part of pipe
        # series, as the cold part is zero milp by construction.
        if heat_network_model_type == "QTH":
            alias_relation = self.alias_relation
        elif heat_network_model_type == NetworkSettings.NETWORK_TYPE_HEAT:
            # There is no proper AliasRelation yet (because there is milp loss in pipes).
            # So we build one, as that is the easiest way to figure out which pipes are
            # connected to each other in series. We do this by making a temporary/shadow
            # discharge (".Q") variable per pipe, as that way we can share the processing
            # logic for determining pipe series with that of QTH models.
            alias_relation = AliasRelation()

            # Look for aliases only in the hot pipes. All cold pipes are zero by convention anyway.
            hot_pipes = self.hot_pipes.copy()
            hot_pipes.extend(self.unrelated_pipes.copy())

            pipes_map = {f"{pipe}.HeatIn.Heat": pipe for pipe in hot_pipes}
            pipes_map.update({f"{pipe}.HeatOut.Heat": pipe for pipe in hot_pipes})

            for p in hot_pipes:
                for port in ["In", "Out"]:
                    heat_port = f"{p}.Heat{port}.Heat"
                    connected = self.alias_relation.aliases(heat_port).intersection(
                        pipes_map.keys()
                    )
                    connected.remove(heat_port)

                    if connected:
                        other_pipe_port = next(iter(connected))
                        if other_pipe_port.endswith(f".Heat{port}.Heat"):
                            sign_prefix = "-"
                        else:
                            sign_prefix = ""
                        other_pipe = pipes_map[other_pipe_port]
                        if f"{other_pipe}.Q" not in alias_relation.canonical_variables:
                            alias_relation.add(f"{p}.Q", f"{sign_prefix}{other_pipe}.Q")

        node_to_node_logical_link_map = {}

        def __get_aliases_asset(asset: str, asset_conn: str) -> List[str]:
            """
            Extracts a list of aliases for a specific variable.

            Args:
                asset: asset name
                asset_conn: asset variable string including port excluding property

            Returns:
                List of aliases for a port property connection
            """
            aliases = [
                x
                for x in self.alias_relation.aliases(f"{asset_conn}.{prop}")
                if not x.startswith(asset) and x.endswith(f".{prop}")
            ]
            return aliases

        for n in [*nodes, *busses, *gas_nodes]:
            n_connections = [ens_params[f"{n}.n"] for ens_params in parameters]

            if len(set(n_connections)) > 1:
                raise Exception(
                    "Nodes and busses cannot have differing number of connections per "
                    "ensemble member"
                )

            n_connections = n_connections[0]

            # Note that we do this based on temperature, because discharge may
            # be an alias of yet some other further away connected pipe.
            if n in nodes:
                node_connections[n] = connected_pipes = {}
            elif n in busses:
                bus_connections[n] = connected_pipes = {}
            elif n in gas_nodes:
                gas_node_connections[n] = connected_pipes = {}

            for i in range(n_connections):
                if n in nodes:
                    cur_port = f"{n}.{heat_network_model_type}Conn[{i + 1}]"
                    prop = (
                        "T"
                        if heat_network_model_type == "QTH"
                        else NetworkSettings.NETWORK_TYPE_HEAT
                    )
                    prop_h = "H"
                    in_suffix = ".QTHIn.T" if heat_network_model_type == "QTH" else ".HeatIn.Heat"
                    out_suffix = (
                        ".QTHOut.T" if heat_network_model_type == "QTH" else ".HeatOut.Heat"
                    )
                    in_suffix_h = ".HeatIn.H"
                    out_suffix_h = ".HeatOut.H"
                    node_suffix = ".HeatConn[1].Heat"
                elif n in busses:
                    cur_port = f"{n}.ElectricityConn[{i + 1}]"
                    prop = "Power"
                    prop_h = "V"
                    in_suffix = ".ElectricityIn.Power"
                    out_suffix = ".ElectricityOut.Power"
                    in_suffix_h = ".ElectricityIn.V"
                    out_suffix_h = ".ElectricityOut.V"
                    node_suffix = ".ElectricityConn[1].Power"
                elif n in gas_nodes:
                    # TODO: Ideally a temporary variable would be created to make the connections
                    #  map that is not passed to the problem
                    cur_port = f"{n}.GasConn[{i + 1}]"
                    prop = "Q"
                    prop_h = "H"
                    in_suffix = ".GasIn.Q"
                    out_suffix = ".GasOut.Q"
                    in_suffix_h = ".GasIn.H"
                    out_suffix_h = ".GasOut.H"
                    node_suffix = ".GasConn[1].Q"
                aliases = __get_aliases_asset(n, cur_port)

                if len(aliases) == 0:
                    raise Exception(f"Found no connection to {cur_port}")

                # Here we make a count of the amount of in and out port aliases.
                in_suffix_count = np.sum([1 if x.endswith(in_suffix) else 0 for x in aliases])
                out_suffix_count = np.sum([1 if x.endswith(out_suffix) else 0 for x in aliases])

                # Here we gather the aliases for a property that is equal for all node ports.
                aliases_h = [
                    x
                    for x in self.alias_relation.aliases(f"{cur_port}.{prop_h}")
                    if not x.startswith(n) and x.endswith(f".{prop_h}")
                ]
                pipe_out_port = False
                # We can have multiple aliases, specifically when a pipe is connected to a port the
                # direction of that pipe matters. To determine if the connected alias is a pipe and
                # which direction it has we look for the overlap between the prop and prop_h in the
                # aliases. This means that if a pipe is both in the aliases and in the aliases_h,
                # then that must be the pipe connected to the port of the node.
                for k in range(len(aliases)):
                    pipe_name = aliases[k].split(".")[0]
                    if pipe_name + out_suffix_h in aliases_h:
                        pipe_out_port = True
                        node_connection_direction = NodeConnectionDirection.IN
                    elif pipe_name + in_suffix_h in aliases_h:
                        pipe_out_port = True
                        node_connection_direction = NodeConnectionDirection.OUT

                if pipe_out_port:
                    # This is only for when a pipe is connected to a gas node to determine direction
                    asset_w_orientation = (
                        pipe_name,
                        node_connection_direction,
                    )
                elif out_suffix_count > in_suffix_count:
                    # This is for the case of Non pipe asset is logically linked to a node
                    asset_w_orientation = (
                        aliases[0][: -len(out_suffix)],
                        NodeConnectionDirection.IN,
                    )
                elif out_suffix_count < in_suffix_count:
                    # This is for the case of Non pipe asset is logically linked to a node
                    asset_w_orientation = (
                        aliases[0][: -len(in_suffix)],
                        NodeConnectionDirection.OUT,
                    )
                elif out_suffix_count == in_suffix_count:
                    # This is for the case of logical links between node to node
                    # Note that we cannot determine the direction of node to node logical links, we
                    # therefore, always take the first node with an in port and the second node with
                    # and out port.
                    if n not in list(node_to_node_logical_link_map.values()):
                        node_to_node_logical_link_map[n] = aliases[0][: -len(node_suffix)]
                        asset_w_orientation = (
                            aliases[0][: -len(node_suffix)],
                            NodeConnectionDirection.IN,
                        )
                    else:
                        asset_w_orientation = (
                            aliases[0][: -len(node_suffix)],
                            NodeConnectionDirection.OUT,
                        )
                else:
                    logger.error("connections are not properly matched")

                connected_pipes[i] = asset_w_orientation

        canonical_pipe_qs = {p: alias_relation.canonical_signed(f"{p}.Q") for p in pipes}
        # Move sign from canonical to alias
        canonical_pipe_qs = {(p, d): c for p, (c, d) in canonical_pipe_qs.items()}
        # Reverse the dictionary from `Dict[alias, canonical]` to `Dict[canonical, Set[alias]]`
        pipe_sets = {}
        for a, c in canonical_pipe_qs.items():
            pipe_sets.setdefault(c, []).append(a)

        pipe_series_with_orientation = list(pipe_sets.values())

        # Check that all pipes in the series have the same orientation
        pipe_series = []
        for ps in pipe_series_with_orientation:
            if not len({orientation for _, orientation in ps}) == 1:
                raise Exception(f"Pipes in series {ps} do not all have the same orientation")
            pipe_series.append([name for name, _ in ps])

        def __get_asset_orientation(in_suffix: str, out_suffix: str) -> Tuple[str, str]:
            """
            Extracts the orientation for the connecting pipes to the asset.
            Args:
                in_suffix: string consisting of inport and property variable name
                out_suffix: string consisting of outport and property variable name

            Returns:
            """
            if aliases[0].endswith(out_suffix):
                asset_w_orientation = (
                    aliases[0][: -len(out_suffix)],
                    NodeConnectionDirection.IN,
                )
            else:
                asset_w_orientation = (
                    aliases[0][: -len(in_suffix)],
                    NodeConnectionDirection.OUT,
                )
            return asset_w_orientation

        def __get_asset_orientation_heat_storages() -> Tuple[str, str]:
            """
            Extracts the orientation of heat storage assets.
            """
            in_suffix = ".QTHIn.T" if heat_network_model_type == "QTH" else ".HeatIn.Heat"
            out_suffix = ".QTHOut.T" if heat_network_model_type == "QTH" else ".HeatOut.Heat"
            return __get_asset_orientation(in_suffix, out_suffix)

        def __get_asset_orientation_source_sink() -> Tuple[str, str]:
            """
            Extracts the orientation of source and sink (demand) assets.
            """
            in_suffix = f".{network_type}In.{prop}"
            out_suffix = f".{network_type}Out.{prop}"
            return __get_asset_orientation(in_suffix, out_suffix)

        def __get_aliases_asset_non_node(asset: str, asset_conn: str) -> List[str]:
            """
            The aliases of non node assets
            """
            aliases = __get_aliases_asset(asset, asset_conn)
            if len(aliases) > 1:
                raise Exception(f"More than one connection to {asset_conn}")
            return aliases

        buffer_connections = {}

        for b in buffers:
            buffer_connections[b] = []

            for k in ["In", "Out"]:
                b_conn = f"{b}.{heat_network_model_type}{k}"
                prop = (
                    "T" if heat_network_model_type == "QTH" else NetworkSettings.NETWORK_TYPE_HEAT
                )
                aliases = __get_aliases_asset_non_node(b, b_conn)
                if len(aliases) == 0:
                    raise Exception(f"Found no connection to {b_conn}")
                asset_w_orientation = __get_asset_orientation_heat_storages()

                assert asset_w_orientation[0] in pipes_set

                # TODO: this check can no longer be based on is_hot_pipe or is_cold_pipe,
                # because the pipes might be drawn separately and then do not have the "related"
                # attribute in esdl.
                # if k == "In":
                #     assert self.is_hot_pipe(asset_w_orientation[0])
                # else:
                #     assert self.is_cold_pipe(asset_w_orientation[0])

                buffer_connections[b].append(asset_w_orientation)

            buffer_connections[b] = tuple(buffer_connections[b])

        ates_connections = {}

        for a in atess:
            ates_connections[a] = []

            for k in ["In", "Out"]:
                a_conn = f"{a}.{heat_network_model_type}{k}"
                prop = (
                    "T" if heat_network_model_type == "QTH" else NetworkSettings.NETWORK_TYPE_HEAT
                )
                aliases = __get_aliases_asset_non_node(a, a_conn)
                if len(aliases) == 0:
                    raise Exception(f"Found no connection to {a_conn}")

                asset_w_orientation = __get_asset_orientation_heat_storages()

                assert asset_w_orientation[0] in pipes_set

                # TODO: this check can no longer be based on is_hot_pipe or is_cold_pipe,
                # because the pipes might be drawn separately and then do not have the "related"
                # attribute in esdl.
                # if k == "Out":
                #     assert self.is_cold_pipe(asset_w_orientation[0])
                # else:
                #     assert self.is_hot_pipe(asset_w_orientation[0])

                ates_connections[a].append(asset_w_orientation)

            ates_connections[a] = tuple(ates_connections[a])

        demand_connections = {}

        for a in demands:
            if a in components.get("heat_demand", []):
                network_type = NetworkSettings.NETWORK_TYPE_HEAT
                prop = NetworkSettings.NETWORK_TYPE_HEAT
            elif a in components.get("electricity_demand", []):
                network_type = "Electricity"
                prop = "Power"
            elif a in components.get("gas_demand", []):
                network_type = NetworkSettings.NETWORK_TYPE_GAS
                prop = "H"
            else:
                logger.error(f"{a} cannot be modelled with heat, gas or electricity")
            a_conn = f"{a}.{network_type}In"

            aliases = __get_aliases_asset_non_node(a, a_conn)
            if len(aliases) == 0:
                # the connection was a logical link to a node
                continue

            asset_w_orientation = __get_asset_orientation_source_sink()

            if asset_w_orientation[0] in [
                *components.get("heat_pipe", []),
                *components.get("gas_pipe", []),
                *components.get("electricity_cable", []),
            ]:
                demand_connections[a] = asset_w_orientation

        source_connections = {}

        for a in sources:
            if a in components.get("heat_source", []):
                network_type = NetworkSettings.NETWORK_TYPE_HEAT
                prop = NetworkSettings.NETWORK_TYPE_HEAT
            elif a in components.get("electricity_source", []):
                network_type = "Electricity"
                prop = "Power"
            elif a in components.get("gas_source", []):
                network_type = NetworkSettings.NETWORK_TYPE_GAS
                prop = "H"
            else:
                logger.error(f"{a} cannot be modelled with heat, gas or electricity")
            a_conn = f"{a}.{network_type}Out"

            aliases = __get_aliases_asset_non_node(a, a_conn)
            if len(aliases) == 0:
                # the connection was a logical link to a node
                continue

            asset_w_orientation = __get_asset_orientation_source_sink()

            if asset_w_orientation[0] in [
                *components.get("heat_pipe", []),
                *components.get("gas_pipe", []),
                *components.get("electricity_cable", []),
            ]:
                source_connections[a] = asset_w_orientation

        self.__topology = Topology(
            node_connections,
            gas_node_connections,
            pipe_series,
            buffer_connections,
            ates_connections,
            bus_connections,
            demand_connections,
            source_connections,
        )

        super().pre()

    @property
    def energy_system_components(self) -> Dict[str, Set[str]]:
        """
        This method return a dict with the milp network assets ordered per asset type.
        """
        if self.__hn_component_types is None:
            # Create the dictionary once after that it will be available
            string_parameters = self.string_parameters(0)

            # Find the components in model, detection by string
            # (name.component_type: type)
            component_types = sorted(
                {v for k, v in string_parameters.items() if "network_type" not in k}
            )

            components = {}
            for c in component_types:
                components[c] = sorted(
                    {k.split(".")[0] for k, v in string_parameters.items() if v == c}
                )

            self.__hn_component_types = components

        return self.__hn_component_types

    @property
    def energy_system_components_commodity(self) -> Dict[str, Set[str]]:
        """
        This method returns a dict with the milp network assets and their commodity type.
        """
        if self.__commodity_types is None:
            # Create the dictionary once after that it will be available
            string_parameters = self.string_parameters(0)

            # Find the components in model, detection by string
            # (name.component_type: type)
            commodity_types = {
                k.split(".")[0]: v for k, v in string_parameters.items() if "network_type" in k
            }

            self.__commodity_types = commodity_types

        return self.__commodity_types

    @property
    def energy_system_topology(self) -> Topology:
        """
        This method returns the topology object of the milp network. Which contains specific assets
        with directions on the ports that are needed in the constraints.
        """
        return self.__topology

    @property
    def hot_to_cold_pipe_map(self) -> Dict:
        """
        This function return a dictionary of hot pipe names mapped to cold pipe names.
        """
        return self.__hot_cold_pipe_relations

    @property
    def cold_to_hot_pipe_map(self) -> Dict:
        """
        This function return a dictionary of cold pipe names mapped to hot pipe names.
        """
        return dict(
            zip(self.__hot_cold_pipe_relations.values(), self.__hot_cold_pipe_relations.keys())
        )

    def hot_cold_pipe_relations(self):
        pipes = self.energy_system_components.get("heat_pipe", [])
        for pipe in pipes:
            related = False
            # test if hot_pipe
            if self.is_hot_pipe(pipe):
                cold_pipe = self.hot_to_cold_pipe(pipe)
                if cold_pipe in pipes:
                    related = True
                    if pipe not in self.__hot_cold_pipe_relations.keys():
                        self.__hot_cold_pipe_relations[pipe] = cold_pipe
            elif self.is_cold_pipe(pipe):
                hot_pipe = self.cold_to_hot_pipe(pipe)
                if hot_pipe in pipes:
                    related = True
                    if hot_pipe not in self.__hot_cold_pipe_relations.keys():
                        self.__hot_cold_pipe_relations[hot_pipe] = pipe
            if not related and pipe not in self.__unrelated_pipes:
                self.__unrelated_pipes.append(pipe)

    @property
    def unrelated_pipes(self) -> List[str]:
        """This function return a list of pipe names of all the pipes that don't have a related
        cold/hot pipe."""
        return self.__unrelated_pipes
